import json
from PIL import Image, ImageFile
import os.path as osp

ImageFile.LOAD_TRUNCATED_IMAGES = True

from io import BytesIO
from typing import Iterable
from torch.utils.data import Dataset, ConcatDataset
import torch


class BaseDataset(Dataset):

    def __init__(self, vis_processor, text_processor, vis_root, anno_path):

        self.vis_root = vis_root
        # if isinstance(anno_path, tuple) or isinstance(anno_path, list):
        #     anno_path = anno_path[0]
        self.anno_path = anno_path

        self.vis_processor = vis_processor
        self.text_processor = text_processor

        self.samples = self.init_samples()
        self.reader = self.init_reader()

        print('total {} {} samples'.format(self.__len__(), self.__class__.__name__))

        for idx in range(10):
            self.__getitem__(idx)

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, index):
        raise NotImplementedError

    def init_samples(self):
        # read annotation from ceph
        if self.anno_path.startswith('cluster'):
            from petrel_client.client import Client
            client = Client("~/petreloss.conf")
            samples = json.loads(client.get(self.anno_path))
        else:
            samples = json.load(open(self.anno_path, 'r'))
        return samples

    def init_reader(self):
        if self.vis_root.startswith('cluster'):
            from petrel_client.client import Client
            client = Client("~/petreloss.conf")
            reader = {'type': 'PetrelReader', 'body': client.get}
        else:
            reader = {'type': 'LocalReader', 'body': Image.open}
        return reader

    def _read_image(self, sample, image_key="image"):
        img_file = sample[image_key]
        image_path = osp.join(self.vis_root, img_file)
        image = self.reader['body'](image_path)
        if isinstance(image, bytes):
            bytes_stream = BytesIO(image)
            image = Image.open(bytes_stream)
        image = image.convert("RGB")
        return image

    def collater(self, samples):
        image_list, question_list, answer_list = [], [], []

        for sample in samples:
            image_list.append(sample["image"])
            question_list.append(sample["text_input"])
            answer_list.append(sample["text_output"])

        return {
            "image": torch.stack(image_list, dim=0),
            "text_input": question_list,
            "text_output": answer_list,
            "data_type": "vqa",
        }


class ConcatDataset(ConcatDataset):
    def __init__(self, datasets: Iterable[Dataset]) -> None:
        super().__init__(datasets)

    def collater(self, samples):
        # TODO For now only supports datasets with same underlying collater implementations

        all_keys = set()
        for s in samples:
            all_keys.update(s)

        shared_keys = all_keys
        for s in samples:
            shared_keys = shared_keys & set(s.keys())

        samples_shared_keys = []
        for s in samples:
            samples_shared_keys.append({k: s[k] for k in s.keys() if k in shared_keys})

        return self.datasets[0].collater(samples_shared_keys)



from doclayout_yolo.engine.model import Model
from doclayout_yolo.nn.tasks import YOLOv10DetectionModel
from .val import YOLOv10DetectionValidator
from .predict import YOLOv10DetectionPredictor
from .train import YOLOv10DetectionTrainer

from huggingface_hub import PyTorchModelHubMixin

class YOLOv10(Model, PyTorchModelHubMixin, repo_url="https://github.com/opendatalab/DocLayout-YOLO", pipeline_tag="object-detection", license="agpl-3.0"):

    def __init__(self, model="yolov10n.pt", task=None, verbose=False):
        super().__init__(model=model, task=task, verbose=verbose)

    @property
    def task_map(self):
        """Map head to model, trainer, validator, and predictor classes."""
        return {
            "detect": {
                "model": YOLOv10DetectionModel,
                "trainer": YOLOv10DetectionTrainer,
                "validator": YOLOv10DetectionValidator,
                "predictor": YOLOv10DetectionPredictor,
            },
        }


import torch
from .base_dataset import BaseDataset
import os.path as osp
import glob
from io import BytesIO
from PIL import Image


class Im2LatexDataset(BaseDataset):

    def init_samples(self):
        samples = []
        for vis_root, anno_path in zip(self.vis_root, self.anno_path):
            images = [path.replace('\\', '/') for path in glob.glob(osp.join(vis_root, '*.png'))]
            indices = [int(osp.basename(img).split('.')[0]) for img in images]

            eqs = open(anno_path, 'r').read().split('\n')
            eqs = [eqs[_] for _ in indices]

            for i, e in zip(images, eqs):
                samples.append({"image": i, "equation": e, "vis_root": vis_root})
        return samples

    def __getitem__(self, index):
        ann = self.samples[index]
        try:
            image = self.vis_processor(self._read_image(ann))
        except Exception:
            return self[(index + 1) % len(self)]
        if image is None:
            return self[(index + 1) % len(self)]
        equation = ann["equation"]
        return {"image": image, "text_input": equation, "id": index}

    def _read_image(self, sample, image_key="image"):
        img_file = sample[image_key]
        vis_root = sample["vis_root"]
        image_path = osp.join(vis_root, img_file)
        image = self.reader['body'](image_path)
        if isinstance(image, bytes):
            bytes_stream = BytesIO(image)
            image = Image.open(bytes_stream)
        image = image.convert("RGB")
        return image

    def init_reader(self):
        if not isinstance(self.vis_root, str):
            vis_root = self.vis_root[0]
        else:
            vis_root = self.vis_root
        if vis_root.startswith('cluster'):
            from petrel_client.client import Client
            client = Client("~/petreloss.conf")
            reader = {'type': 'PetrelReader', 'body': client.get}
        else:
            reader = {'type': 'LocalReader', 'body': Image.open}
        return reader

    def collater(self, samples):
        image_list, question_list, id_list = [], [], []

        for sample in samples:
            image_list.append(sample["image"])
            question_list.append(sample["text_input"])
            id_list.append(sample["id"])

        return {
            "image": torch.stack(image_list, dim=0),
            "text_input": question_list,
            "id": id_list
        }



from unimernet.common.registry import registry
from omegaconf import OmegaConf
import albumentations as alb
from albumentations.pytorch import ToTensorV2
from unimernet.processors.base_processor import BaseProcessor
import numpy as np
import cv2
from PIL import Image, ImageOps
from torchvision.transforms.functional import resize
import random


class FormulaImageBaseProcessor(BaseProcessor):

    def __init__(self, image_size):
        super(FormulaImageBaseProcessor, self).__init__()
        self.input_size = [int(_) for _ in image_size]
        assert len(self.input_size) == 2

    @staticmethod
    def crop_margin(img: Image.Image) -> Image.Image:
        data = np.array(img.convert("L"))
        data = data.astype(np.uint8)
        max_val = data.max()
        min_val = data.min()
        if max_val == min_val:
            return img
        data = (data - min_val) / (max_val - min_val) * 255
        gray = 255 * (data < 200).astype(np.uint8)

        coords = cv2.findNonZero(gray)  # Find all non-zero points (text)
        a, b, w, h = cv2.boundingRect(coords)  # Find minimum spanning bounding box
        return img.crop((a, b, w + a, h + b))

    def prepare_input(self, img: Image.Image, random_padding: bool = False):
        """
        Convert PIL Image to tensor according to specified input_size after following steps below:
            - resize
            - rotate (if align_long_axis is True and image is not aligned longer axis with canvas)
            - pad
        """
        if img is None:
            return
        # crop margins
        try:
            img = self.crop_margin(img.convert("RGB"))
        except OSError:
            # might throw an error for broken files
            return

        if img.height == 0 or img.width == 0:
            return

        img = resize(img, min(self.input_size))
        img.thumbnail((self.input_size[1], self.input_size[0]))
        delta_width = self.input_size[1] - img.width
        delta_height = self.input_size[0] - img.height
        if random_padding:
            pad_width = np.random.randint(low=0, high=delta_width + 1)
            pad_height = np.random.randint(low=0, high=delta_height + 1)
        else:
            pad_width = delta_width // 2
            pad_height = delta_height // 2
        padding = (
            pad_width,
            pad_height,
            delta_width - pad_width,
            delta_height - pad_height,
        )
        return ImageOps.expand(img, padding)


@registry.register_processor("formula_image_train")
class FormulaImageTrainProcessor(FormulaImageBaseProcessor):
    def __init__(self, image_size=384):
        super().__init__(image_size)

        # Import weather-related augmentations only when initializing this class
        from unimernet.processors.formula_processor_helper.nougat import Bitmap, Dilation, Erosion
        from unimernet.processors.formula_processor_helper.weather import Fog, Frost, Snow, Rain, Shadow

        self.transform = alb.Compose(
            [
                alb.Compose(
                    [
                        Bitmap(p=0.05),
                        alb.OneOf([Fog(), Frost(), Snow(), Rain(), Shadow()], p=0.2),
                        alb.OneOf([Erosion((2, 3)), Dilation((2, 3))], p=0.2),
                        alb.ShiftScaleRotate(shift_limit=0, scale_limit=(-.15, 0), rotate_limit=1, border_mode=0,
                                             interpolation=3,
                                             value=[255, 255, 255],
                                             p=1),
                        alb.GridDistortion(distort_limit=0.1, border_mode=0, interpolation=3, value=[255, 255, 255],
                                           p=.5)],
                    p=.15),
                # alb.InvertImg(p=.15),
                alb.RGBShift(r_shift_limit=15, g_shift_limit=15, b_shift_limit=15, p=0.3),
                alb.GaussNoise(10, p=.2),
                alb.RandomBrightnessContrast(.05, (-.2, 0), True, p=0.2),
                alb.ImageCompression(95, p=.3),
                alb.ToGray(always_apply=True),
                alb.Normalize((0.7931, 0.7931, 0.7931), (0.1738, 0.1738, 0.1738)),
                # alb.Sharpen()
                ToTensorV2(),
            ]
        )

    def __call__(self, item):
        img = self.prepare_input(item, random_padding=True)
        if img is None:
            return img
        return self.transform(image=np.array(img))['image'][:1]

    @classmethod
    def from_config(cls, cfg=None):
        if cfg is None:
            cfg = OmegaConf.create()

        image_size = cfg.get("image_size", [384, 384])

        return cls(
            image_size=image_size,
        )


@registry.register_processor("formula_image_multi_scale_train")
class FormulaImageMultiScaleTrainProcessor(FormulaImageTrainProcessor):
    def __init__(self, all_scales):
        for i, scales in enumerate(all_scales):
            all_scales[i] = [int(_) for _ in scales]
        super(FormulaImageMultiScaleTrainProcessor, self).__init__(all_scales[0])
        self.all_scales = all_scales

    @classmethod
    def from_config(cls, cfg=None):
        if cfg is None:
            cfg = OmegaConf.create()

        all_scales = cfg.get("all_scales", [[384, 384]])
        return cls(
            all_scales=all_scales
        )

    def reset_scale(self):
        self.input_size = random.choice(self.all_scales)


@registry.register_processor("formula_image_eval")
class FormulaImageEvalProcessor(FormulaImageBaseProcessor):
    def __init__(self, image_size):
        super().__init__(image_size)

        self.transform = alb.Compose(
            [
                alb.ToGray(always_apply=True),
                alb.Normalize((0.7931, 0.7931, 0.7931), (0.1738, 0.1738, 0.1738)),
                # alb.Sharpen()
                ToTensorV2(),
            ]
        )

    def __call__(self, item):
        image = self.prepare_input(item)
        return self.transform(image=np.array(image))['image'][:1]

    @classmethod
    def from_config(cls, cfg=None):
        if cfg is None:
            cfg = OmegaConf.create()

        image_size = cfg.get("image_size", [384, 384])

        return cls(image_size=image_size)



"""
 Copyright (c) 2022, salesforce.com, inc.
 All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause
 For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""

import datetime
import logging
import os
import time

import torch
import torch.distributed as dist
import webdataset as wds
from unimernet.common.dist_utils import download_cached_file, is_main_process, main_process
from unimernet.common.registry import registry
from unimernet.common.utils import is_url
from unimernet.datasets.data_utils import reorg_datasets_by_split
from unimernet.runners.runner_base import RunnerBase
from torch.utils.data.dataset import ChainDataset


@registry.register_runner("runner_iter")
class RunnerIter(RunnerBase):
    """
    Run training based on the number of iterations. This is common when
    the training dataset size is large. Underhood logic is similar to
    epoch-based training by considering every #iters_per_inner_epoch as an
    inner epoch.

    In iter-based runner, after every #iters_per_inner_epoch steps, we

        1) do a validation epoch;
        2) schedule the learning rate;
        3) save the checkpoint.

    We refer every #iters_per_inner_epoch steps as an inner epoch.
    """

    def __init__(self, cfg, task, model, datasets, job_id):
        super().__init__(cfg, task, model, datasets, job_id)

        self.start_iters = 0

        self.max_iters = int(self.config.run_cfg.get("max_iters", -1))
        assert self.max_iters > 0, "max_iters must be greater than 0."

        self.iters_per_inner_epoch = int(
            self.config.run_cfg.get("iters_per_inner_epoch", -1)
        )
        assert (
                self.iters_per_inner_epoch > 0
        ), "iters_per_inner_epoch must be greater than 0."

    @property
    def max_epoch(self):
        return int(self.max_iters / self.iters_per_inner_epoch)

    @property
    def cur_epoch(self):
        try:
            return self.train_loader.epoch
        except AttributeError:
            # pipeline data (e.g. LAION) is streaming, have no concept of epoch
            return 0

    def _progress(self, cur_iters):
        return "{}_iters={}".format(self.cur_epoch, cur_iters)

    def train(self):
        start_time = time.time()
        best_agg_metric = 0
        best_iters = 0

        self.log_config()

        # resume from checkpoint if specified
        if not self.evaluate_only and self.resume_ckpt_path is not None:
            self._load_checkpoint(self.resume_ckpt_path)
        cur_epoch = 0
        for start_iters in range(
                self.start_iters, self.max_iters, self.iters_per_inner_epoch
        ):
            end_iters = start_iters + self.iters_per_inner_epoch

            # training phase
            if not self.evaluate_only:
                logging.info(
                    "Start training, max_iters={}, in total {} inner epochs.".format(
                        self.max_iters, int(self.max_iters / self.iters_per_inner_epoch)
                    )
                )

                train_stats = self.train_iters(self.cur_epoch, start_iters)
                self.log_stats(split_name="train", stats=train_stats)

            # evaluation phase
            if len(self.valid_splits) > 0:
                for split_name in self.valid_splits:
                    logging.info("Evaluating on {}.".format(split_name))

                    val_log = self.eval_epoch(
                        split_name=split_name, cur_epoch=self._progress(end_iters)
                    )
                    if val_log is not None:
                        if is_main_process():
                            assert (
                                    "agg_metrics" in val_log
                            ), "No agg_metrics found in validation log."

                            agg_metrics = val_log["agg_metrics"]
                            if agg_metrics > best_agg_metric and split_name == "eval":
                                best_iters, best_agg_metric = end_iters, agg_metrics

                                self._save_checkpoint(end_iters, is_best=True)
                            val_log.update({"best_iters": best_iters})
                            self.log_stats(val_log, split_name)
                            # print evaluation metric
                            print(f"bleu:{val_log['bleu']:.6f}, edit_distance:{val_log['edit_distance']:.6f}, token_accuracy:{val_log['token_accuracy']:.6f} ")
                            print("="*80)

            if self.evaluate_only:
                break
            if self.milestone and cur_epoch + 1 in self.milestone:
                self._save_checkpoint(cur_epoch)
            self._save_checkpoint(end_iters, latest=True)
            dist.barrier()
            cur_epoch += 1

        # testing phase
        self.evaluate(cur_epoch=self.cur_epoch)

        total_time = time.time() - start_time
        total_time_str = str(datetime.timedelta(seconds=int(total_time)))
        logging.info("Training time {}".format(total_time_str))

    def train_iters(self, epoch, start_iters):
        # train by iterations
        self.model.train()

        return self.task.train_iters(
            epoch=epoch,
            start_iters=start_iters,
            iters_per_inner_epoch=self.iters_per_inner_epoch,
            model=self.model,
            data_loader=self.train_loader,
            optimizer=self.optimizer,
            scaler=self.scaler,
            lr_scheduler=self.lr_scheduler,
            cuda_enabled=self.cuda_enabled,
            log_freq=self.log_freq,
            accum_grad_iters=self.accum_grad_iters,
        )

    @main_process
    def _save_checkpoint(self, cur_iters, is_best=False, latest=False):
        # only save the params requires gradient
        assert not (is_best and latest), "You can't set 'is_best' and 'latest' the same time."
        unwrapped_model = self.unwrap_dist_model(self.model)
        param_grad_dic = {
            k: v.requires_grad for (k, v) in unwrapped_model.named_parameters()
        }

        state_dict = unwrapped_model.state_dict()
        for k in list(state_dict.keys()):
            if k in param_grad_dic.keys() and not param_grad_dic[k]:
                del state_dict[k]

        save_obj = {
            "model": state_dict,
            "optimizer": self.optimizer.state_dict(),
            "config": self.config.to_dict(),
            "scaler": self.scaler.state_dict() if self.scaler else None,
            "iters": cur_iters,
        }
        if is_best:
            save_to = os.path.join(
                self.output_dir,
                "checkpoint_{}.pth".format("best"),
            )
        elif latest:
            save_to = os.path.join(
                self.output_dir,
                "checkpoint_{}.pth".format("latest"),
            )
        else:
            save_to = os.path.join(
                self.output_dir,
                "checkpoint_{}.pth".format(cur_iters),
            )
        logging.info("Saving checkpoint at iters {} to {}.".format(cur_iters, save_to))
        torch.save(save_obj, save_to)

    def _load_checkpoint(self, url_or_filename):
        """
        Resume from a checkpoint.
        """
        if is_url(url_or_filename):
            cached_file = download_cached_file(
                url_or_filename, check_hash=False, progress=True
            )
            checkpoint = torch.load(cached_file, map_location=self.device)
        elif os.path.isfile(url_or_filename):
            checkpoint = torch.load(url_or_filename, map_location=self.device)
        else:
            raise RuntimeError("checkpoint url or path is invalid")

        state_dict = checkpoint["model"]
        self.unwrap_dist_model(self.model).load_state_dict(state_dict)

        self.optimizer.load_state_dict(checkpoint["optimizer"])
        if self.scaler and "scaler" in checkpoint:
            self.scaler.load_state_dict(checkpoint["scaler"])

        self.start_iters = checkpoint["iters"] + 1
        logging.info("Resume checkpoint from {}".format(url_or_filename))

    @property
    def dataloaders(self) -> dict:
        """
        A property to get and create dataloaders by split just in need.

        If no train_dataset_ratio is provided, concatenate map-style datasets and
        chain wds.DataPipe datasets separately. Training set becomes a tuple
        (ConcatDataset, ChainDataset), both are optional but at least one of them is
        required. The resultant ConcatDataset and ChainDataset will be sampled evenly.

        If train_dataset_ratio is provided, create a MultiIterLoader to sample
        each dataset by ratios during training.

        Currently do not support multiple datasets for validation and test.

        Returns:
            dict: {split_name: (tuples of) dataloader}
        """
        if self._dataloaders is None:
            # reoganize datasets by split and concatenate/chain if necessary

            self.datasets = reorg_datasets_by_split(self.datasets)
            # to keep the same structure as return value of concat_datasets
            self.datasets = {
                k: v[0] if len(v) == 1 else v for k, v in self.datasets.items()
            }

            # print dataset statistics after concatenation/chaining
            for split_name in self.datasets:
                if isinstance(self.datasets[split_name], tuple) or isinstance(
                        self.datasets[split_name], list
                ):
                    # mixed wds.DataPipeline and torch.utils.data.Dataset
                    num_records = sum(
                        [
                            len(d)
                            if not type(d) in [wds.DataPipeline, ChainDataset]
                            else 0
                            for d in self.datasets[split_name]
                        ]
                    )

                else:
                    try:
                        # a single map-style dataset
                        num_records = len(self.datasets[split_name])
                    except TypeError:
                        # a single wds.DataPipeline or ChainDataset
                        num_records = -1
                        logging.info(
                            "Only a single wds.DataPipeline dataset, no __len__ attribute."
                        )

                if num_records >= 0:
                    logging.info(
                        "Loaded {} records for {} split from the dataset.".format(
                            num_records, split_name
                        )
                    )

            # create dataloaders
            split_names = sorted(self.datasets.keys())

            datasets = [self.datasets[split] for split in split_names]
            is_trains = [split in self.train_splits for split in split_names]

            batch_sizes = [
                self.config.run_cfg.batch_size_train
                if split == "train"
                else self.config.run_cfg.batch_size_eval
                for split in split_names
            ]

            collate_fns = []
            for dataset in datasets:
                if isinstance(dataset, tuple) or isinstance(dataset, list):
                    collate_fns.append([getattr(d, "collater", None) for d in dataset])
                else:
                    collate_fns.append(getattr(dataset, "collater", None))

            dataloaders = self.create_loaders(
                datasets=datasets,
                num_workers=self.config.run_cfg.num_workers,
                batch_sizes=batch_sizes,
                is_trains=is_trains,
                collate_fns=collate_fns,
            )

            self._dataloaders = {k: v for k, v in zip(split_names, dataloaders)}

        return self._dataloaders



from doclayout_yolo.models.yolo.detect import DetectionValidator
from doclayout_yolo.utils import ops
import torch

import pdb

class YOLOv10DetectionValidator(DetectionValidator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.args.save_json |= self.is_coco

    def postprocess(self, preds, conf=None):
        if isinstance(preds, dict):
            preds = preds["one2one"]

        if isinstance(preds, (list, tuple)):
            preds = preds[0]
        
        preds = preds.transpose(-1, -2)
        boxes, scores, labels = ops.v10postprocess(preds, self.args.max_det, self.nc)
        bboxes = ops.xywh2xyxy(boxes)
        # return torch.cat([bboxes, scores.unsqueeze(-1), labels.unsqueeze(-1)], dim=-1)
        
        preds = torch.cat([bboxes, scores.unsqueeze(-1), labels.unsqueeze(-1)], dim=-1)
        if preds.shape[-1] == 6 and conf is not None:  # end-to-end model (BNC, i.e. 1,300,6)
            preds = [pred[pred[:, 4] > conf] for pred in preds]
        return preds


import torch
import torch.nn.functional as F
from unimernet.common.registry import registry
from unimernet.models.blip2_models.blip2 import Blip2Base
from unimernet.models.unimernet.encoder_decoder import DonutEncoderDecoder, DonutTokenizer


@registry.register_model("unimernet")
class UniMERModel(Blip2Base):
    """
    Nougat model for formula recognition.
    Supported model types:
        - default
    Usage:
        >>> from unimernet.models import load_model
        >>> model = load_model("unimernet", "default")
    """

    PRETRAINED_MODEL_CONFIG_DICT = {
        "default": "configs/models/unimernet_base.yaml",
        "unimernet": "configs/models/unimernet_base.yaml",
    }

    def __init__(
            self,
            *,
            model_name,
            model_config,
            tokenizer_name,
            tokenizer_config,
    ):
        super().__init__()

        self.tokenizer = DonutTokenizer(tokenizer_config.path)
        self.model = DonutEncoderDecoder(
            model_config.model_name,
            num_tokens=len(self.tokenizer),
            bos_token_id=self.tokenizer.bos_token_id,
            pad_token_id=self.tokenizer.pad_token_id,
            eos_token_id=self.tokenizer.eos_token_id,
        )
        self.max_seq_len = model_config.max_seq_len
        self.tokenizer.max_seq_len = self.max_seq_len

    def forward(self, samples):
        image, text = samples["image"], samples["text_input"]

        text_inputs = self.tokenizer.tokenize(text).to(image.device)
        count_gt = self._get_count_gt(text, image.device)
        tgt_seq, tgt_mask = text_inputs["input_ids"], text_inputs["attention_mask"]
        with self.maybe_autocast():
            loss = self.model(
                pixel_values=image,
                decoder_input_ids=tgt_seq,
                decoder_attention_mask=tgt_mask,
                decoder_count_gt=count_gt,
            )
        return {"loss": loss}

    def _get_count_gt(self, text, device):
        labels = self.tokenizer.tokenize(text, max_length=1536)["input_ids"].to(device)
        mask = labels != self.tokenizer.pad_token_id
        one_hot_labels = F.one_hot(labels, num_classes=self.tokenizer.tokenizer.vocab_size) * mask.unsqueeze(-1)
        count_gt = torch.sum(one_hot_labels, dim=1)
        return count_gt # (bs, vocab_size)

    @torch.no_grad()
    def generate(
            self,
            samples,
            temperature: float = 0.2,
            do_sample: bool = False,
            top_p: float = 0.95,
            **kwargs
    ):

        image = samples["image"]
        with self.maybe_autocast():
            outputs = self.model.generate(
                pixel_values=image,
                temperature=temperature,
                max_new_tokens=self.max_seq_len,
                decoder_start_token_id=self.tokenizer.tokenizer.bos_token_id,
                # decoder_end_token_id=self.tokenizer.tokenizer.eos_token_id,
                do_sample=do_sample,
                top_p=top_p,
                **kwargs
            )
        pred_tokens = self.tokenizer.detokenize(outputs)
        pred_str = self.tokenizer.token2str(outputs)
        return {"pred_tokens": pred_tokens, "pred_str": pred_str, "pred_ids": outputs}

    @classmethod
    def from_config(cls, cfg):

        model_name = cfg.get("model_name")
        model_config = cfg.get("model_config")
        tokenizer_name = cfg.get("tokenizer_name")
        tokenizer_config = cfg.get("tokenizer_config")

        model = cls(
            model_name=model_name,
            model_config=model_config,
            tokenizer_name=tokenizer_name,
            tokenizer_config=tokenizer_config
        )

        model.load_checkpoint_from_config(cfg)

        return model



import torch
from .formula import Im2LatexDataset


class MultiScaleIm2LatexDataset(Im2LatexDataset):

    def __getitem__(self, index):
        ann = self.samples[index]
        try:
            pil_image = self._read_image(ann)
            image = self.vis_processor(pil_image)
        except Exception:
            return self[(index + 1) % len(self)]
        if image is None:
            return self[(index + 1) % len(self)]
        equation = ann["equation"]
        return {"image": image, "text_input": equation, "id": index, "raw_image": pil_image}

    def collater(self, samples):
        self.vis_processor.reset_scale()
        image_list, question_list, id_list = [], [], []

        for sample in samples:
            image_list.append(self.vis_processor(sample["raw_image"]))
            question_list.append(sample["text_input"])
            id_list.append(sample["id"])

        return {
            "image": torch.stack(image_list, dim=0),
            "text_input": question_list,
            "id": id_list
        }



"""
Model downloader for FreeP2W
Automatically downloads UniMERNet model on first run
"""

import os
import sys
import urllib.request
from pathlib import Path
import zipfile
import shutil


def get_model_dir():
    """Get the directory where models should be stored"""
    # For installed package, use user's home directory
    home = Path.home()
    model_dir = home / ".freep2w" / "weights"
    model_dir.mkdir(parents=True, exist_ok=True)
    return model_dir


def download_file(url, dest_path, description="Downloading"):
    """Download file with progress bar"""
    print(f"[INFO] {description}...")

    # Convert Path to string for urllib
    dest_path = str(dest_path)

    def progress_hook(count, block_size, total_size):
        if total_size > 0:
            percent = int(count * block_size * 100 / total_size)
            sys.stdout.write(f"\r[INFO] 下载进度: {percent}%")
            sys.stdout.flush()

    try:
        urllib.request.urlretrieve(url, dest_path, progress_hook)
        print()  # New line after download
        print(f"[INFO] 下载完成: {dest_path}")
    except Exception as e:
        print(f"\n[错误] 下载失败: {e}")
        raise


def check_and_download_models():
    """
    Check if models exist, download if needed
    Returns: (yolo_model_path, unimernet_cfg_path)
    """
    model_dir = get_model_dir()

    # Check YOLO model
    yolo_model = model_dir / "doclayout_yolo_docstructbench_imgsz1024.pt"
    if not yolo_model.exists():
        try:
            YOLO_MODEL_URL = "https://github.com/zstar1003/FreeP2W/releases/download/v0.0.1/doclayout_yolo_docstructbench_imgsz1024.pt"
            download_file(YOLO_MODEL_URL, yolo_model, "下载 YOLO 模型")
        except Exception as e:
            print(f"[警告] 从 GitHub 下载失败: {e}")
            print("[INFO] 尝试使用包内的模型文件...")
            # Fallback to package's weights directory
            package_model = Path(__file__).parent / "weights" / "doclayout_yolo_docstructbench_imgsz1024.pt"
            if package_model.exists():
                shutil.copy(package_model, yolo_model)
                print(f"[OK] 已从包目录复制模型到: {yolo_model}")
            else:
                print("[ERROR] 包内也找不到模型文件")
                raise FileNotFoundError(f"YOLO 模型不存在: {yolo_model}")

    # Check UniMERNet model
    unimernet_dir = model_dir / "unimernet_small"
    if not unimernet_dir.exists():
        try:
            download_unimernet_model(unimernet_dir)
        except Exception as e:
            print(f"[警告] UniMERNet 模型下载失败: {e}")
            print("[INFO] 将使用项目根目录的 unimernet 模块")
            # Note: The unimernet module is already in sys.path via package installation

    # Check config file
    demo_yaml = model_dir / "demo.yaml"
    unimernet_dir = model_dir / "unimernet_small"

    # Always regenerate config with correct paths
    default_yaml = Path(__file__).parent / "demo.yaml"
    if default_yaml.exists():
        # Read default config
        with open(default_yaml, 'r', encoding='utf-8') as f:
            config_content = f.read()

        # Replace relative paths with absolute paths
        config_content = config_content.replace(
            './weights/unimernet_small',
            str(unimernet_dir).replace('\\', '/')
        )

        # Write updated config
        with open(demo_yaml, 'w', encoding='utf-8') as f:
            f.write(config_content)

        print(f"[OK] 已生成配置文件: {demo_yaml}")
        print(f"[OK] UniMERNet 模型路径: {unimernet_dir}")

    return str(yolo_model), str(demo_yaml)


def download_unimernet_model(dest_dir):
    print("[INFO] 自动下载 UniMERNet 模型...")
    url = "https://github.com/zstar1003/FreeP2W/releases/download/v0.0.1/unimernet_small.zip"
    zip_path = dest_dir.parent / "unimernet_small.zip"
    download_file(url, zip_path, "下载 UniMERNet 模型")

    # Extract to the specific unimernet_small directory
    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        zip_ref.extractall(dest_dir)  # Extract directly to dest_dir (weights/unimernet_small/)

    zip_path.unlink()
    print(f"[OK] UniMERNet 模型已解压到: {dest_dir}")


if __name__ == "__main__":
    yolo_path, cfg_path = check_and_download_models()
    print(f"YOLO 模型路径: {yolo_path}")
    print(f"配置文件路径: {cfg_path}")


# Ultralytics YOLO 🚀, AGPL-3.0 license

from .rtdetr import RTDETR
from .sam import SAM
from .yolo import YOLO, YOLOWorld
from .yolov10 import YOLOv10

__all__ = "YOLO", "RTDETR", "SAM", "YOLOWorld", "YOLOv10"  # allow simpler import



"""
 Copyright (c) 2022, salesforce.com, inc.
 All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause
 For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""

import gzip
import logging
import os
import random as rnd
import tarfile
import zipfile

import webdataset as wds
import numpy as np
import torch
from torch.utils.data.dataset import IterableDataset, ChainDataset
from unimernet.common.registry import registry
from unimernet.datasets.datasets.base_dataset import ConcatDataset
from tqdm import tqdm

MAX_INT = registry.get("MAX_INT")


def apply_to_sample(f, sample):
    if len(sample) == 0:
        return {}

    def _apply(x):
        if torch.is_tensor(x):
            return f(x)
        elif isinstance(x, dict):
            return {key: _apply(value) for key, value in x.items()}
        elif isinstance(x, list):
            return [_apply(x) for x in x]
        else:
            return x

    return _apply(sample)


def move_to_cuda(sample):
    def _move_to_cuda(tensor):
        return tensor.cuda()

    return apply_to_sample(_move_to_cuda, sample)


def prepare_sample(samples, cuda_enabled=True):
    if cuda_enabled:
        samples = move_to_cuda(samples)

    # TODO fp16 support

    return samples


def reorg_datasets_by_split(datasets):
    """
    Organizes datasets by split.

    Args:
        datasets: dict of torch.utils.data.Dataset objects by name.

    Returns:
        Dict of datasets by split {split_name: List[Datasets]}.
    """
    # if len(datasets) == 1:
    #     return datasets[list(datasets.keys())[0]]
    # else:
    reorg_datasets = dict()

    # reorganize by split
    for _, dataset in datasets.items():
        for split_name, dataset_split in dataset.items():
            if split_name not in reorg_datasets:
                reorg_datasets[split_name] = [dataset_split]
            else:
                reorg_datasets[split_name].append(dataset_split)

    return reorg_datasets


def concat_datasets(datasets):
    """
    Concatenates multiple datasets into a single dataset.

    It supports may-style datasets and DataPipeline from WebDataset. Currently, does not support
    generic IterableDataset because it requires creating separate samplers.

    Now only supports conctenating training datasets and assuming validation and testing
    have only a single dataset. This is because metrics should not be computed on the concatenated
    datasets.

    Args:
        datasets: dict of torch.utils.data.Dataset objects by split.

    Returns:
        Dict of concatenated datasets by split, "train" is the concatenation of multiple datasets,
        "val" and "test" remain the same.

        If the input training datasets contain both map-style and DataPipeline datasets, returns
        a tuple, where the first element is a concatenated map-style dataset and the second
        element is a chained DataPipeline dataset.

    """
    # concatenate datasets in the same split
    for split_name in datasets:
        if split_name != "train":
            assert (
                len(datasets[split_name]) == 1
            ), "Do not support multiple {} datasets.".format(split_name)
            datasets[split_name] = datasets[split_name][0]
        else:
            iterable_datasets, map_datasets = [], []
            for dataset in datasets[split_name]:
                if isinstance(dataset, wds.DataPipeline):
                    logging.info(
                        "Dataset {} is IterableDataset, can't be concatenated.".format(
                            dataset
                        )
                    )
                    iterable_datasets.append(dataset)
                elif isinstance(dataset, IterableDataset):
                    raise NotImplementedError(
                        "Do not support concatenation of generic IterableDataset."
                    )
                else:
                    map_datasets.append(dataset)

            # if len(iterable_datasets) > 0:
            # concatenate map-style datasets and iterable-style datasets separately
            chained_datasets = (
                ChainDataset(iterable_datasets) if len(iterable_datasets) > 0 else None
            )
            concat_datasets = (
                ConcatDataset(map_datasets) if len(map_datasets) > 0 else None
            )

            train_datasets = concat_datasets, chained_datasets
            train_datasets = tuple([x for x in train_datasets if x is not None])
            train_datasets = (
                train_datasets[0] if len(train_datasets) == 1 else train_datasets
            )

            datasets[split_name] = train_datasets

    return datasets


def extract_archive(from_path, to_path=None, overwrite=False):
    """Extract archive.

    Args:
        from_path: the path of the archive.
        to_path: the root path of the extracted files (directory of from_path)
        overwrite: overwrite existing files (False)

    Returns:
        List of paths to extracted files even if not overwritten.

    Examples:
        >>> url = 'http://www.quest.dcs.shef.ac.uk/wmt16_files_mmt/validation.tar.gz'
        >>> from_path = './validation.tar.gz'
        >>> to_path = './'
        >>> torchtext.utils.download_from_url(url, from_path)
        >>> torchtext.utils.extract_archive(from_path, to_path)
        >>> ['.data/val.de', '.data/val.en']
        >>> torchtext.utils.download_from_url(url, from_path)
        >>> torchtext.utils.extract_archive(from_path, to_path)
        >>> ['.data/val.de', '.data/val.en']

    """

    if to_path is None:
        to_path = os.path.dirname(from_path)

    if from_path.endswith((".tar.gz", ".tgz")):
        logging.info("Opening tar file {} to {}.".format(from_path, to_path))
        with tarfile.open(from_path, "r") as tar:
            files = []
            for file_ in tqdm(tar):
                file_path = os.path.join(to_path, file_.name)
                if file_.isfile():
                    files.append(file_path)
                    if os.path.exists(file_path):
                        logging.info("{} already extracted.".format(file_path))
                        if not overwrite:
                            continue
                tar.extract(file_, to_path)
            logging.info("Finished extracting tar file {}.".format(from_path))
            return files

    elif from_path.endswith(".zip"):
        assert zipfile.is_zipfile(from_path), from_path
        logging.info("Opening zip file {} to {}.".format(from_path, to_path))
        with zipfile.ZipFile(from_path, "r") as zfile:
            files = []
            for file_ in tqdm(zfile.namelist()):
                file_path = os.path.join(to_path, file_)
                files.append(file_path)
                if os.path.exists(file_path):
                    logging.info("{} already extracted.".format(file_path))
                    if not overwrite:
                        continue
                zfile.extract(file_, to_path)
        files = [f for f in files if os.path.isfile(f)]
        logging.info("Finished extracting zip file {}.".format(from_path))
        return files

    elif from_path.endswith(".gz"):
        logging.info("Opening gz file {} to {}.".format(from_path, to_path))
        default_block_size = 65536
        filename = from_path[:-3]
        files = [filename]
        with gzip.open(from_path, "rb") as gzfile, open(filename, "wb") as d_file:
            while True:
                block = gzfile.read(default_block_size)
                if not block:
                    break
                else:
                    d_file.write(block)
            d_file.write(block)
        logging.info("Finished extracting gz file {}.".format(from_path))
        return files

    else:
        raise NotImplementedError(
            "We currently only support tar.gz, .tgz, .gz and zip achives."
        )


def save_frames_grid(img_array, out_path):
    import torch
    from PIL import Image
    from torchvision.utils import make_grid

    if len(img_array.shape) == 3:
        img_array = img_array.unsqueeze(0)
    elif len(img_array.shape) == 5:
        b, t, c, h, w = img_array.shape
        img_array = img_array.view(-1, c, h, w)
    elif len(img_array.shape) == 4:
        pass
    else:
        raise NotImplementedError(
            "Supports only (b,t,c,h,w)-shaped inputs. First two dimensions can be ignored."
        )

    assert img_array.shape[1] == 3, "Exepcting input shape of (H, W, 3), i.e. RGB-only."

    grid = make_grid(img_array)
    ndarr = grid.permute(1, 2, 0).to("cpu", torch.uint8).numpy()

    img = Image.fromarray(ndarr)

    img.save(out_path)



import albumentations as alb
import numpy as np
import cv2


class Erosion(alb.ImageOnlyTransform):
    """
    Apply erosion operation to an image.

    Erosion is a morphological operation that shrinks the white regions in a binary image.

    Args:
        scale (int or tuple/list of int): The scale or range for the size of the erosion kernel.
            If an integer is provided, a square kernel of that size will be used.
            If a tuple or list is provided, it should contain two integers representing the minimum
            and maximum sizes for the erosion kernel.
        always_apply (bool, optional): Whether to always apply this transformation. Default is False.
        p (float, optional): The probability of applying this transformation. Default is 0.5.

    Returns:
        numpy.ndarray: The transformed image.
    """

    def __init__(self, scale, always_apply=False, p=0.5):
        super().__init__(always_apply=always_apply, p=p)
        if type(scale) is tuple or type(scale) is list:
            assert len(scale) == 2
            self.scale = scale
        else:
            self.scale = (scale, scale)

    def apply(self, img, **params):
        kernel = cv2.getStructuringElement(
            cv2.MORPH_ELLIPSE, tuple(np.random.randint(self.scale[0], self.scale[1], 2))
        )
        img = cv2.erode(img, kernel, iterations=1)
        return img


class Dilation(alb.ImageOnlyTransform):
    """
    Apply dilation operation to an image.

    Dilation is a morphological operation that expands the white regions in a binary image.

    Args:
        scale (int or tuple/list of int): The scale or range for the size of the dilation kernel.
            If an integer is provided, a square kernel of that size will be used.
            If a tuple or list is provided, it should contain two integers representing the minimum
            and maximum sizes for the dilation kernel.
        always_apply (bool, optional): Whether to always apply this transformation. Default is False.
        p (float, optional): The probability of applying this transformation. Default is 0.5.

    Returns:
        numpy.ndarray: The transformed image.
    """

    def __init__(self, scale, always_apply=False, p=0.5):
        super().__init__(always_apply=always_apply, p=p)
        if type(scale) is tuple or type(scale) is list:
            assert len(scale) == 2
            self.scale = scale
        else:
            self.scale = (scale, scale)

    def apply(self, img, **params):
        kernel = cv2.getStructuringElement(
            cv2.MORPH_ELLIPSE, tuple(np.random.randint(self.scale[0], self.scale[1], 2))
        )
        img = cv2.dilate(img, kernel, iterations=1)
        return img


class Bitmap(alb.ImageOnlyTransform):
    """
    Apply a bitmap-style transformation to an image.

    This transformation replaces all pixel values below a certain threshold with a specified value.

    Args:
        value (int, optional): The value to replace pixels below the threshold with. Default is 0.
        lower (int, optional): The threshold value below which pixels will be replaced. Default is 200.
        always_apply (bool, optional): Whether to always apply this transformation. Default is False.
        p (float, optional): The probability of applying this transformation. Default is 0.5.

    Returns:
        numpy.ndarray: The transformed image.
    """

    def __init__(self, value=0, lower=200, always_apply=False, p=0.5):
        super().__init__(always_apply=always_apply, p=p)
        self.lower = lower
        self.value = value

    def apply(self, img, **params):
        img = img.copy()
        img[img < self.lower] = self.value
        return img



# Ultralytics YOLO 🚀, AGPL-3.0 license

import getpass
from typing import List

import cv2
import numpy as np
import pandas as pd

from doclayout_yolo.data.augment import LetterBox
from doclayout_yolo.utils import LOGGER as logger
from doclayout_yolo.utils import SETTINGS
from doclayout_yolo.utils.checks import check_requirements
from doclayout_yolo.utils.ops import xyxy2xywh
from doclayout_yolo.utils.plotting import plot_images


def get_table_schema(vector_size):
    """Extracts and returns the schema of a database table."""
    from lancedb.pydantic import LanceModel, Vector

    class Schema(LanceModel):
        im_file: str
        labels: List[str]
        cls: List[int]
        bboxes: List[List[float]]
        masks: List[List[List[int]]]
        keypoints: List[List[List[float]]]
        vector: Vector(vector_size)

    return Schema


def get_sim_index_schema():
    """Returns a LanceModel schema for a database table with specified vector size."""
    from lancedb.pydantic import LanceModel

    class Schema(LanceModel):
        idx: int
        im_file: str
        count: int
        sim_im_files: List[str]

    return Schema


def sanitize_batch(batch, dataset_info):
    """Sanitizes input batch for inference, ensuring correct format and dimensions."""
    batch["cls"] = batch["cls"].flatten().int().tolist()
    box_cls_pair = sorted(zip(batch["bboxes"].tolist(), batch["cls"]), key=lambda x: x[1])
    batch["bboxes"] = [box for box, _ in box_cls_pair]
    batch["cls"] = [cls for _, cls in box_cls_pair]
    batch["labels"] = [dataset_info["names"][i] for i in batch["cls"]]
    batch["masks"] = batch["masks"].tolist() if "masks" in batch else [[[]]]
    batch["keypoints"] = batch["keypoints"].tolist() if "keypoints" in batch else [[[]]]
    return batch


def plot_query_result(similar_set, plot_labels=True):
    """
    Plot images from the similar set.

    Args:
        similar_set (list): Pyarrow or pandas object containing the similar data points
        plot_labels (bool): Whether to plot labels or not
    """
    similar_set = (
        similar_set.to_dict(orient="list") if isinstance(similar_set, pd.DataFrame) else similar_set.to_pydict()
    )
    empty_masks = [[[]]]
    empty_boxes = [[]]
    images = similar_set.get("im_file", [])
    bboxes = similar_set.get("bboxes", []) if similar_set.get("bboxes") is not empty_boxes else []
    masks = similar_set.get("masks") if similar_set.get("masks")[0] != empty_masks else []
    kpts = similar_set.get("keypoints") if similar_set.get("keypoints")[0] != empty_masks else []
    cls = similar_set.get("cls", [])

    plot_size = 640
    imgs, batch_idx, plot_boxes, plot_masks, plot_kpts = [], [], [], [], []
    for i, imf in enumerate(images):
        im = cv2.imread(imf)
        im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
        h, w = im.shape[:2]
        r = min(plot_size / h, plot_size / w)
        imgs.append(LetterBox(plot_size, center=False)(image=im).transpose(2, 0, 1))
        if plot_labels:
            if len(bboxes) > i and len(bboxes[i]) > 0:
                box = np.array(bboxes[i], dtype=np.float32)
                box[:, [0, 2]] *= r
                box[:, [1, 3]] *= r
                plot_boxes.append(box)
            if len(masks) > i and len(masks[i]) > 0:
                mask = np.array(masks[i], dtype=np.uint8)[0]
                plot_masks.append(LetterBox(plot_size, center=False)(image=mask))
            if len(kpts) > i and kpts[i] is not None:
                kpt = np.array(kpts[i], dtype=np.float32)
                kpt[:, :, :2] *= r
                plot_kpts.append(kpt)
        batch_idx.append(np.ones(len(np.array(bboxes[i], dtype=np.float32))) * i)
    imgs = np.stack(imgs, axis=0)
    masks = np.stack(plot_masks, axis=0) if plot_masks else np.zeros(0, dtype=np.uint8)
    kpts = np.concatenate(plot_kpts, axis=0) if plot_kpts else np.zeros((0, 51), dtype=np.float32)
    boxes = xyxy2xywh(np.concatenate(plot_boxes, axis=0)) if plot_boxes else np.zeros(0, dtype=np.float32)
    batch_idx = np.concatenate(batch_idx, axis=0)
    cls = np.concatenate([np.array(c, dtype=np.int32) for c in cls], axis=0)

    return plot_images(
        imgs, batch_idx, cls, bboxes=boxes, masks=masks, kpts=kpts, max_subplots=len(images), save=False, threaded=False
    )


def prompt_sql_query(query):
    """Plots images with optional labels from a similar data set."""
    check_requirements("openai>=1.6.1")
    from openai import OpenAI

    if not SETTINGS["openai_api_key"]:
        logger.warning("OpenAI API key not found in settings. Please enter your API key below.")
        openai_api_key = getpass.getpass("OpenAI API key: ")
        SETTINGS.update({"openai_api_key": openai_api_key})
    openai = OpenAI(api_key=SETTINGS["openai_api_key"])

    messages = [
        {
            "role": "system",
            "content": """
                You are a helpful data scientist proficient in SQL. You need to output exactly one SQL query based on
                the following schema and a user request. You only need to output the format with fixed selection
                statement that selects everything from "'table'", like `SELECT * from 'table'`

                Schema:
                im_file: string not null
                labels: list<item: string> not null
                child 0, item: string
                cls: list<item: int64> not null
                child 0, item: int64
                bboxes: list<item: list<item: double>> not null
                child 0, item: list<item: double>
                    child 0, item: double
                masks: list<item: list<item: list<item: int64>>> not null
                child 0, item: list<item: list<item: int64>>
                    child 0, item: list<item: int64>
                        child 0, item: int64
                keypoints: list<item: list<item: list<item: double>>> not null
                child 0, item: list<item: list<item: double>>
                    child 0, item: list<item: double>
                        child 0, item: double
                vector: fixed_size_list<item: float>[256] not null
                child 0, item: float

                Some details about the schema:
                - the "labels" column contains the string values like 'person' and 'dog' for the respective objects
                    in each image
                - the "cls" column contains the integer values on these classes that map them the labels

                Example of a correct query:
                request - Get all data points that contain 2 or more people and at least one dog
                correct query-
                SELECT * FROM 'table' WHERE  ARRAY_LENGTH(cls) >= 2  AND ARRAY_LENGTH(FILTER(labels, x -> x = 'person')) >= 2  AND ARRAY_LENGTH(FILTER(labels, x -> x = 'dog')) >= 1;
             """,
        },
        {"role": "user", "content": f"{query}"},
    ]

    response = openai.chat.completions.create(model="gpt-3.5-turbo", messages=messages)
    return response.choices[0].message.content



"""
 Copyright (c) 2022, salesforce.com, inc.
 All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause
 For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""

from unimernet.common.registry import registry
from unimernet.tasks.base_task import BaseTask
from unimernet.tasks.unimernet_train import UniMERNet_Train


def setup_task(cfg):
    assert "task" in cfg.run_cfg, "Task name must be provided."

    task_name = cfg.run_cfg.task
    task = registry.get_task_class(task_name).setup_task(cfg=cfg)
    assert task is not None, "Task {} not properly registered.".format(task_name)

    return task


__all__ = [
    "BaseTask",
    "UniMERNet_Train",
]



import logging
from unimernet.common.registry import registry
from unimernet.datasets.builders.base_dataset_builder import BaseDatasetBuilder
from unimernet.datasets.datasets.formula import Im2LatexDataset
from unimernet.datasets.datasets.formula_multi_scale import MultiScaleIm2LatexDataset


@registry.register_builder("formula_rec_train")
class FormulaRecTrainBuilder(BaseDatasetBuilder):
    train_dataset_cls = Im2LatexDataset
    DATASET_CONFIG_DICT = {
        "default": "configs/datasets/formula/formula_train.yaml"
    }
    LOG_INFO = "Formula Recgnition Train"

    def build_datasets(self):
        logging.info(f"Building {self.LOG_INFO} datasets ...")
        self.build_processors()

        build_info = self.config.build_info
        anno_path = build_info.annotation,
        vis_root = build_info.images
        anno_path = [anno_path] if isinstance(anno_path, str) else anno_path
        vis_root = [vis_root] if isinstance(vis_root, str) else vis_root
        datasets = dict()

        # create datasets
        dataset_cls = self.train_dataset_cls
        datasets['train'] = dataset_cls(
            vis_processor=self.vis_processors["train"],
            text_processor=self.text_processors["train"],
            vis_root=vis_root,
            anno_path=anno_path,
        )
        print(datasets['train'][0])

        return datasets


@registry.register_builder("multi_scale_formula_rec_train")
class MultiScaleFormulaRecTrainBuilder(BaseDatasetBuilder):
    train_dataset_cls = MultiScaleIm2LatexDataset
    DATASET_CONFIG_DICT = {
        "default": "configs/datasets/formula/multi_scale_formula_train.yaml"
    }
    LOG_INFO = "Multi Scale Formula Recgnition Train"

    def build_datasets(self):
        logging.info(f"Building {self.LOG_INFO} datasets ...")
        self.build_processors()

        build_info = self.config.build_info
        anno_path = build_info.annotation,
        vis_root = build_info.images

        anno_path = [anno_path] if isinstance(anno_path, str) else anno_path
        vis_root = [vis_root] if isinstance(vis_root, str) else vis_root

        datasets = dict()

        # create datasets
        dataset_cls = self.train_dataset_cls
        datasets['train'] = dataset_cls(
            vis_processor=self.vis_processors["train"],
            text_processor=self.text_processors["train"],
            vis_root=vis_root,
            anno_path=anno_path,
        )
        print(datasets['train'][0])

        return datasets


@registry.register_builder("formula_rec_eval")
class FormulaRecEvalBuilder(BaseDatasetBuilder):
    eval_dataset_cls = Im2LatexDataset
    DATASET_CONFIG_DICT = {
        "default": "configs/datasets/formula/formula_eval.yaml"
    }
    LOG_INFO = "Formula Recgnition Eval"

    def build_datasets(self):
        logging.info(f"Building {self.LOG_INFO} datasets ...")
        self.build_processors()

        build_info = self.config.build_info
        anno_path = build_info.annotation,
        vis_root = build_info.images

        anno_path = [anno_path] if isinstance(anno_path, str) else anno_path
        vis_root = [vis_root] if isinstance(vis_root, str) else vis_root

        datasets = dict()

        # create datasets
        dataset_cls = self.eval_dataset_cls
        datasets['eval'] = dataset_cls(
            vis_processor=self.vis_processors["eval"],
            text_processor=self.text_processors["eval"],
            vis_root=vis_root,
            anno_path=anno_path,
        )
        print(datasets['eval'][0])

        return datasets



"""
FreeP2W - Free PDF to Word Converter
Command-line tool with clean output
"""

import sys
import os
import warnings
import logging
from pathlib import Path

# Add project root to sys.path to ensure local modules are found
# This is necessary when running via entry point
project_root = Path(__file__).parent.parent.absolute()
if str(project_root) not in sys.path:
    sys.path.insert(0, str(project_root))

# 禁用所有警告
warnings.filterwarnings('ignore')

# 配置日志只显示关键信息
logging.basicConfig(
    level=logging.CRITICAL,  # 只显示严重错误
    format='%(message)s'
)

# 禁用第三方库的日志
logging.getLogger('pdf2docx').setLevel(logging.CRITICAL)
logging.getLogger('fitz').setLevel(logging.CRITICAL)
logging.getLogger('PIL').setLevel(logging.CRITICAL)


def print_info(message):
    """打印 [INFO] 信息"""
    print(f"[INFO] {message}")


def print_save(filepath):
    """打印 [保存] 信息"""
    print(f"[保存] 已保存到: {filepath}")


def convert_pdf_to_docx(pdf_path, output_path=None):
    """
    转换 PDF 到 DOCX

    Args:
        pdf_path: PDF 文件路径
        output_path: 输出 DOCX 路径（可选）
    """

    # 自动检查并下载模型
    yolo_path = None
    cfg_path = None
    try:
        from .model_downloader import check_and_download_models
        print_info("正在检查模型文件...")
        yolo_path, cfg_path = check_and_download_models()
        print(f"[DEBUG] YOLO路径: {yolo_path}")
        print(f"[DEBUG] 配置路径: {cfg_path}")
    except Exception as e:
        print(f"[警告] 模型检查失败: {e}")
        print("[INFO] 将使用默认路径继续...")
        import traceback
        traceback.print_exc()
    
    # 验证输入文件
    if not os.path.exists(pdf_path):
        print(f"[错误] PDF 文件不存在: {pdf_path}")
        return False

    # 确定输出路径
    if output_path is None:
        base_name = os.path.splitext(os.path.basename(pdf_path))[0]
        output_path = f"{base_name}_converted.docx"

    try:
        # 导入转换器（延迟导入以加快启动速度）
        print_info("正在加载转换器...")

        # 重定向标准输出到空
        import io
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        sys.stdout = io.StringIO()
        sys.stderr = io.StringIO()

        try:
            from .hybrid_converter import HybridConverter
            from pdf2docx import Converter

            # 恢复输出
            sys.stdout = old_stdout
            sys.stderr = old_stderr

            print_info("开始转换 PDF...")

            # 转换 PDF（使用下载的模型路径）
            sys.stdout = io.StringIO()
            sys.stderr = io.StringIO()

            # 使用下载的模型路径，如果没有则使用默认相对路径
            if yolo_path and cfg_path:
                converter = HybridConverter(
                    yolo_model_path=yolo_path,
                    unimernet_cfg_path=cfg_path
                )
            else:
                # Fallback: 使用项目目录的相对路径
                converter = HybridConverter(
                    yolo_model_path="weights/doclayout_yolo_docstructbench_imgsz1024.pt",
                    unimernet_cfg_path="demo.yaml"
                )

            converter.convert(pdf_path=pdf_path, docx_path=output_path)

            # converter = Converter(pdf_path)
            # converter.convert(output_path)

            sys.stdout = old_stdout
            sys.stderr = old_stderr

            print_info("PDF 转换完成")

            print_save(output_path)
            return True

        except Exception as e:
            sys.stdout = old_stdout
            sys.stderr = old_stderr
            print(f"[错误] 转换失败: {e}")
            return False

    except Exception as e:
        print(f"[错误] 加载失败: {e}")
        return False


def main():
    """主函数"""
    import argparse

    parser = argparse.ArgumentParser(
        description='FreeP2W - Free PDF to Word Converter',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
示例:
  freep2w input.pdf                    # 转换为 input_converted.docx
  freep2w input.pdf -o output.docx     # 指定输出文件名
  freep2w test.pdf -o result.docx      # 完整示例
        """
    )

    parser.add_argument('input', help='输入 PDF 文件路径')
    parser.add_argument('-o', '--output', help='输出 DOCX 文件路径（可选）')
    parser.add_argument('-v', '--version', action='version', version='FreeP2W 1.0')

    args = parser.parse_args()

    # 执行转换
    success = convert_pdf_to_docx(args.input, args.output)

    if success:
        sys.exit(0)
    else:
        sys.exit(1)


if __name__ == "__main__":
    main()



import torch
import torch.nn as nn

from . import hybrid
from . import vit
from . import transformer


class Model(nn.Module):
    def __init__(self, encoder, decoder, args):
        super().__init__()
        self.encoder = encoder
        self.decoder = decoder
        self.args = args

    def data_parallel(self, x: torch.Tensor, device_ids, output_device=None, **kwargs):
        if not device_ids or len(device_ids) == 1:
            return self(x, **kwargs)
        if output_device is None:
            output_device = device_ids[0]
        replicas = nn.parallel.replicate(self, device_ids)
        inputs = nn.parallel.scatter(x, device_ids)  # Slices tensors into approximately equal chunks and distributes them across given GPUs.
        kwargs = nn.parallel.scatter(kwargs, device_ids)  # Duplicates references to objects that are not tensors.
        replicas = replicas[:len(inputs)]
        kwargs = kwargs[:len(inputs)]
        outputs = nn.parallel.parallel_apply(replicas, inputs, kwargs)
        return nn.parallel.gather(outputs, output_device).mean()

    def forward(self, x: torch.Tensor, tgt_seq: torch.Tensor,  **kwargs):
        encoded = self.encoder(x)
        out = self.decoder(tgt_seq, context=encoded, **kwargs)
        return out

    @torch.no_grad()
    def generate(self, x: torch.Tensor, temperature: float = 0.25):
        return self.decoder.generate((torch.LongTensor([self.args.bos_token]*len(x))[:, None]).to(x.device), self.args.max_seq_len,
                                     eos_token=self.args.eos_token, context=self.encoder(x), temperature=temperature)


def get_model(args):
    if args.encoder_structure.lower() == 'vit':
        encoder = vit.get_encoder(args)
    elif args.encoder_structure.lower() == 'hybrid':
        encoder = hybrid.get_encoder(args)
    else:
        raise NotImplementedError('Encoder structure "%s" not supported.' % args.encoder_structure)
    decoder = transformer.get_decoder(args)
    encoder.to(args.device)
    decoder.to(args.device)
    model = Model(encoder, decoder, args)
    if args.wandb:
        import wandb
        wandb.watch(model)

    return model


from doclayout_yolo.models.yolo.detect import DetectionPredictor
import torch
from doclayout_yolo.utils import ops
from doclayout_yolo.engine.results import Results


class YOLOv10DetectionPredictor(DetectionPredictor):
    def postprocess(self, preds, img, orig_imgs):
        if isinstance(preds, dict):
            preds = preds["one2one"]

        if isinstance(preds, (list, tuple)):
            preds = preds[0]

        if preds.shape[-1] == 6:
            pass
        else:
            preds = preds.transpose(-1, -2)
            bboxes, scores, labels = ops.v10postprocess(preds, self.args.max_det, preds.shape[-1]-4)
            bboxes = ops.xywh2xyxy(bboxes)
            preds = torch.cat([bboxes, scores.unsqueeze(-1), labels.unsqueeze(-1)], dim=-1)

        mask = preds[..., 4] > self.args.conf
        if self.args.classes is not None:
            mask = mask & (preds[..., 5:6] == torch.tensor(self.args.classes, device=preds.device).unsqueeze(0)).any(2)
        
        preds = [p[mask[idx]] for idx, p in enumerate(preds)]

        if not isinstance(orig_imgs, list):  # input images are a torch.Tensor, not a list
            orig_imgs = ops.convert_torch2numpy_batch(orig_imgs)

        results = []
        for i, pred in enumerate(preds):
            orig_img = orig_imgs[i]
            pred[:, :4] = ops.scale_boxes(img.shape[2:], pred[:, :4], orig_img.shape)
            img_path = self.batch[0][i]
            results.append(Results(orig_img, path=img_path, names=self.model.names, boxes=pred))
        return results



from typing import Dict, Union, Optional, List

from torch import TensorType
from transformers import DonutImageProcessor, DonutProcessor
from transformers.image_processing_utils import BatchFeature
from transformers.image_transforms import pad
from transformers.image_utils import PILImageResampling, ImageInput, ChannelDimension, make_list_of_images, \
    valid_images, to_numpy_array, is_scaled_image, get_image_size
import numpy as np
import PIL
import logging

logger = logging.getLogger()

IMAGE_STD = [0.229, 0.224, 0.225]
IMAGE_MEAN = [0.485, 0.456, 0.406]


class VariableDonutImageProcessor(DonutImageProcessor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def numpy_resize(self, image: np.ndarray, size, resample):
        image = PIL.Image.fromarray(image)
        resized = self.pil_resize(image, size, resample)
        resized = np.array(resized, dtype=np.uint8)
        resized_image = resized.transpose(2, 0, 1)

        return resized_image

    def pil_resize(self, image: PIL.Image.Image, size, resample):
        width, height = image.size
        max_width, max_height = size["width"], size["height"]
        if width != max_width or height != max_height:
            # Shrink to fit within dimensions
            width_scale = max_width / width
            height_scale = max_height / height
            scale = min(width_scale, height_scale)

            new_width = min(int(width * scale), max_width)
            new_height = min(int(height * scale), max_height)

            image = image.resize((new_width, new_height), resample)

        image.thumbnail((max_width, max_height), resample)

        assert image.width <= max_width and image.height <= max_height

        return image

    def process_inner(self, images: List[List], train=False):
        # This will be in list of lists format, with height x width x channel
        assert isinstance(images[0], (list, np.ndarray))

        # convert list of lists format to array
        if isinstance(images[0], list):
            # numpy unit8 needed for augmentation
            np_images = [np.array(img, dtype=np.uint8) for img in images]
        else:
            np_images = [img.astype(np.uint8) for img in images]

        assert np_images[0].shape[2] == 3  # RGB input images, channel dim last

        # This also applies the right channel dim format, to channel x height x width
        np_images = [self.numpy_resize(img, self.max_size, self.resample) for img in np_images]
        assert np_images[0].shape[0] == 3  # RGB input images, channel dim first

        # Convert to float32 for rescale/normalize
        np_images = [img.astype(np.float32) for img in np_images]

        # Pads with 255 (whitespace)
        # Pad to max size to improve performance
        max_size = self.max_size
        np_images = [
            self.pad_image(
                image=image,
                size=max_size,
                random_padding=train,  # Change amount of padding randomly during training
                input_data_format=ChannelDimension.FIRST,
                pad_value=255.0
            )
            for image in np_images
        ]

        # Rescale and normalize
        np_images = [
            self.rescale(img, scale=self.rescale_factor, input_data_format=ChannelDimension.FIRST)
            for img in np_images
        ]
        np_images = [
            self.normalize(img, mean=self.image_mean, std=self.image_std, input_data_format=ChannelDimension.FIRST)
            for img in np_images
        ]

        return np_images

    def preprocess(
            self,
            images: ImageInput,
            do_resize: bool = None,
            size: Dict[str, int] = None,
            resample: PILImageResampling = None,
            do_thumbnail: bool = None,
            do_align_long_axis: bool = None,
            do_pad: bool = None,
            random_padding: bool = False,
            do_rescale: bool = None,
            rescale_factor: float = None,
            do_normalize: bool = None,
            image_mean: Optional[Union[float, List[float]]] = None,
            image_std: Optional[Union[float, List[float]]] = None,
            return_tensors: Optional[Union[str, TensorType]] = None,
            data_format: Optional[ChannelDimension] = ChannelDimension.FIRST,
            input_data_format: Optional[Union[str, ChannelDimension]] = None,
            **kwargs,
    ) -> PIL.Image.Image:
        images = make_list_of_images(images)

        if not valid_images(images):
            raise ValueError(
                "Invalid image type. Must be of type PIL.Image.Image, numpy.ndarray, "
                "torch.Tensor, tf.Tensor or jax.ndarray."
            )

        # Convert to numpy for later processing steps
        images = [to_numpy_array(image) for image in images]

        images = self.process_inner(images, train=False)

        data = {"pixel_values": images}
        return BatchFeature(data=data, tensor_type=return_tensors)

    def pad_image(
            self,
            image: np.ndarray,
            size: Dict[str, int],
            random_padding: bool = False,
            data_format: Optional[Union[str, ChannelDimension]] = None,
            input_data_format: Optional[Union[str, ChannelDimension]] = None,
            pad_value: float = 0.0,
    ) -> np.ndarray:
        output_height, output_width = size["height"], size["width"]
        input_height, input_width = get_image_size(image, channel_dim=input_data_format)

        delta_width = output_width - input_width
        delta_height = output_height - input_height

        assert delta_width >= 0 and delta_height >= 0

        if random_padding:
            pad_top = np.random.randint(low=0, high=delta_height + 1)
            pad_left = np.random.randint(low=0, high=delta_width + 1)
        else:
            pad_top = delta_height // 2
            pad_left = delta_width // 2

        pad_bottom = delta_height - pad_top
        pad_right = delta_width - pad_left

        padding = ((pad_top, pad_bottom), (pad_left, pad_right))
        return pad(image, padding, data_format=data_format, input_data_format=input_data_format,
                   constant_values=pad_value)


class VariableDonutProcessor(DonutProcessor):
    def __init__(self, image_processor=None, tokenizer=None, train=False, **kwargs):
        if image_processor is None:
            raise ValueError("You need to specify an `image_processor`.")
        if tokenizer is None:
            raise ValueError("You need to specify a `tokenizer`.")

        super().__init__(image_processor, tokenizer)
        self.current_processor = self.image_processor
        self._in_target_context_manager = False
        self.train = train

    def __call__(self, *args, **kwargs):
        # For backward compatibility
        if self._in_target_context_manager:
            return self.current_processor(*args, **kwargs)

        images = kwargs.pop("images", None)
        text = kwargs.pop("text", None)
        if len(args) > 0:
            images = args[0]
            args = args[1:]

        if images is None:
            raise ValueError("You need to specify images to process.")

        inputs = self.image_processor(images, *args, **kwargs)
        return inputs



import math
from functools import partial
from typing import Any, Callable, List, Optional, Tuple

from .conv import Conv
from .block import CIB

import torch
from torch import nn, Tensor
import torch.nn.functional as F

class DilatedBlock(nn.Module):
    """Standard bottleneck with dilated convolution."""

    def __init__(self, c, dilation, k, fuse="sum", shortcut=True):
        """Initializes a bottleneck module with given input/output channels, shortcut option, group, kernels, and
        expansion.
        """
        super().__init__()
        self.dilation = dilation
        self.k = k
        self.cv2 = Conv(c, c, k=1, s=1)
        self.add = shortcut
        
        self.fuse = fuse
        if fuse == "glu":
            self.conv_gating = Conv(c*len(self.dilation), c*len(self.dilation), k=1, s=1, g=c*len(self.dilation))
            self.conv1x1 = Conv(c*len(self.dilation), c, k=1, s=1, g=c)
        elif fuse == "sum":
            self.conv1x1 = Conv(c, c, k=1, s=1, g=c)
            
        self.dcv = Conv(c, c, k=self.k, s=1)

    def dilated_conv(self, x, dilation):
        act = self.dcv.act
        bn = self.dcv.bn
        weight = self.dcv.conv.weight
        padding = dilation * (self.k//2)
        return act(bn(F.conv2d(x, weight, stride=1, padding=padding, dilation=dilation)))
    
    def forward(self, x):
        """'forward()' applies the YOLO FPN to input data."""
        dx = [self.dilated_conv(x, d) for d in self.dilation]
        dx = [self.cv2(_dx) for _dx in dx]
        if self.fuse == "glu":
            dx = torch.cat(dx, dim=1)
            G = torch.sigmoid(self.conv_gating(dx))
            dx = dx * G  # Element-wise multiplication
            dx = self.conv1x1(dx)
        elif self.fuse == "sum":
            dx = [_dx.unsqueeze(0) for _dx in dx]
            dx = torch.cat(dx, dim=0)
            dx = torch.sum(dx, dim=0)
            dx = self.conv1x1(dx)
            
        return x + dx if self.add else dx
        

class DilatedBottleneck(nn.Module):
    """Standard bottleneck with dilated convolution."""

    def __init__(self, c1, c2, shortcut=True, dilation=[1,2,3], block_k=3, fuse="sum", g=1, k=(3, 3), e=0.5):
        """Initializes a bottleneck module with given input/output channels, shortcut option, group, kernels, and
        expansion.
        """
        super().__init__()
        
        c_ = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, c_, k[0], 1)
        self.cv2 = Conv(c_, c2, k[1], 1, g=g)
        
        self.dilated_block = DilatedBlock(c_, dilation, block_k, fuse)
        self.add = shortcut and c1 == c2

    def forward(self, x):
        """'forward()' applies the YOLO FPN to input data."""
        return x + self.cv2(self.dilated_block(self.cv1(x))) if self.add else self.cv2(self.dilated_block(self.cv1(x)))
    
class G2L_CRM(nn.Module):
    """Faster Implementation of CSP Bottleneck with 2 convolutions."""

    def __init__(self, c1, c2, n=1, shortcut=False, use_dilated=False, dilation=[1,2,3], block_k=3, fuse="sum", g=1, e=0.5):
        """Initialize CSP bottleneck layer with two convolutions with arguments ch_in, ch_out, number, shortcut, groups,
        expansion.
        """
        super().__init__()
        self.c = int(c2 * e)  # hidden channels
        self.cv1 = Conv(c1, 2 * self.c, 1, 1)
        self.cv2 = Conv((2 + n) * self.c, c2, 1)  # optional act=FReLU(c2)
        if use_dilated:
            self.m = nn.ModuleList(DilatedBottleneck(
                self.c, 
                self.c, 
                shortcut, 
                dilation, 
                block_k, 
                fuse, 
                g, 
                k=((3, 3), (3, 3)), 
                e=1.0) for _ in range(n)
            )
        else:
            self.m = nn.ModuleList(CIB(
                self.c, 
                self.c, 
                shortcut, 
                e=1.0) for _ in range(n)
            )

    def forward(self, x):
        """Forward pass through C2f layer."""
        y = list(self.cv1(x).chunk(2, 1))
        for m in self.m:
            y.append(m(y[-1]))
        return self.cv2(torch.cat(y, 1))

    def forward_split(self, x):
        """Forward pass using split() instead of chunk()."""
        y = list(self.cv1(x).split((self.c, self.c), 1))
        for m in self.m:
            y.append(m(y[-1]))
        return self.cv2(torch.cat(y, 1))


"""
Hybrid PDF to DOCX Converter
Combines DocLayout-YOLO (for images and formulas) + pdf2docx (for text/tables)

Strategy:
1. Use DocLayout-YOLO to detect figure and formula regions
2. Use UniMERNet to recognize formulas and convert to MathML
3. Use pdf2docx for text, tables, and other content
4. Replace pdf2docx's image extraction with YOLO-detected regions
"""

import os
import torch
from doclayout_yolo import YOLOv10
from pdf2docx import Converter
import fitz


class YOLORegionDetector:
    """DocLayout-YOLO region detector for pdf2docx integration (figures + formulas)"""

    def __init__(self, model_path="weights/doclayout_yolo_docstructbench_imgsz1024.pt",
                 imgsz=1024, conf=0.25, dpi=150):
        """
        Initialize YOLO detector

        Args:
            model_path: Path to YOLO model
            imgsz: Image size for detection
            conf: Confidence threshold
            dpi: DPI for PDF rendering
        """
        print(f"[YOLO] Loading DocLayout-YOLO model...")

        # Auto select device
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        print(f"[YOLO] Using device: {self.device}")

        # Load model
        self.model = YOLOv10(model_path)
        self.imgsz = imgsz
        self.conf = conf
        self.dpi = dpi

        # Cache for detected regions (page_num -> {'figures': [...], 'formulas': [...]})
        self.detected_regions = {}

        print(f"[YOLO] Model loaded successfully\n")

    def detect_page_regions(self, page, page_num):
        """
        Detect figure and formula regions in a PDF page

        Args:
            page: fitz.Page object
            page_num: Page number (0-indexed)

        Returns:
            dict: {'figures': [regions], 'formulas': [regions]}
                  Each region: {'bbox': [x0,y0,x1,y1], 'confidence': float}
        """
        # Check cache
        if page_num in self.detected_regions:
            return self.detected_regions[page_num]

        # Convert page to image
        mat = fitz.Matrix(self.dpi / 72, self.dpi / 72)
        pix = page.get_pixmap(matrix=mat)

        # Save temp image
        temp_img = f"_temp_yolo_page_{page_num}.png"
        pix.save(temp_img)

        # Run YOLO detection
        results = self.model.predict(
            temp_img,
            imgsz=self.imgsz,
            conf=self.conf,
            device=self.device,
            verbose=False  # Suppress YOLO output
        )

        # Clean up temp file
        if os.path.exists(temp_img):
            os.remove(temp_img)

        # Parse results
        figures = []
        formulas = []

        if len(results) > 0 and hasattr(results[0], 'boxes'):
            for box in results[0].boxes:
                # Get class
                cls_id = int(box.cls[0].cpu().numpy())
                class_name = self.model.names[cls_id]

                # Get bbox in image coordinates
                x1, y1, x2, y2 = box.xyxy[0].cpu().numpy()

                # Convert to PDF coordinates
                scale_x = page.rect.width / pix.width
                scale_y = page.rect.height / pix.height

                pdf_bbox = [
                    float(x1 * scale_x),
                    float(y1 * scale_y),
                    float(x2 * scale_x),
                    float(y2 * scale_y)
                ]

                region_data = {
                    'bbox': pdf_bbox,
                    'confidence': float(box.conf[0].cpu().numpy())
                }

                # Categorize by type
                if class_name.lower() in ['figure', 'picture']:
                    figures.append(region_data)
                elif class_name.lower() in ['isolate_formula', 'formula']:
                    formulas.append(region_data)

        # Cache results
        regions = {'figures': figures, 'formulas': formulas}
        self.detected_regions[page_num] = regions

        if figures or formulas:
            print(f"[YOLO] Page {page_num + 1}: Detected {len(figures)} figures, {len(formulas)} formulas")

        return regions


class HybridConverter:
    """
    Hybrid PDF to DOCX Converter
    Uses DocLayout-YOLO for images and formulas, pdf2docx for everything else
    """

    def __init__(self, yolo_model_path="weights/doclayout_yolo_docstructbench_imgsz1024.pt",
                 unimernet_cfg_path="demo.yaml"):
        """
        Initialize hybrid converter

        Args:
            yolo_model_path: Path to DocLayout-YOLO model
            unimernet_cfg_path: Path to UniMERNet config file (default: demo.yaml in project root)
        """
        self.yolo_detector = YOLORegionDetector(model_path=yolo_model_path)
        self.unimernet_cfg_path = unimernet_cfg_path
        self.formula_recognizer = None  # Lazy load UniMERNet
        self.formula_processor = None

    def _load_unimernet(self):
        """Lazy load UniMERNet model"""
        if self.formula_recognizer is not None:
            return

        print("[UniMERNet] Loading formula recognition model...")

        import sys
        import os

        # Get project root directory (where hybrid_converter.py is located)
        project_root = os.path.dirname(os.path.abspath(__file__))

        # Add project root to path if not already there
        if project_root not in sys.path:
            sys.path.insert(0, project_root)

        try:
            import argparse
            import unimernet.tasks as tasks
            from unimernet.common.config import Config
            from unimernet.processors import load_processor

            # Setup model - use absolute path for config
            cfg_path = os.path.join(project_root, self.unimernet_cfg_path)
            if not os.path.exists(cfg_path):
                raise FileNotFoundError(f"Config file not found: {cfg_path}")

            args = argparse.Namespace(cfg_path=cfg_path, options=None)
            cfg = Config(args)

            task = tasks.setup_task(cfg)
            device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

            # Build and load model
            self.formula_recognizer = task.build_model(cfg).to(device)
            self.formula_recognizer.eval()

            # Load processor
            self.formula_processor = load_processor(
                "formula_image_eval",
                cfg.config.datasets.formula_rec_eval.vis_processor.eval,
            )

            print(f"[UniMERNet] Model loaded on {device}\n")

        except ImportError as e:
            print(f"[Error] Failed to import UniMERNet modules: {e}")
            print(f"[Error] Make sure 'unimernet' folder is in: {project_root}")
            raise
        except Exception as e:
            print(f"[Error] Failed to load UniMERNet model: {e}")
            raise

    def _recognize_formula(self, page, bbox):
        """
        Recognize formula from PDF page region

        Args:
            page: fitz.Page object
            bbox: [x0, y0, x1, y1] coordinates

        Returns:
            str: LaTeX formula string
        """
        # Ensure UniMERNet is loaded
        self._load_unimernet()

        # Extract formula region as image
        rect = fitz.Rect(bbox)
        mat = fitz.Matrix(3.0, 3.0)  # Higher resolution for formula recognition
        pix = page.get_pixmap(matrix=mat, clip=rect)

        # Convert to PIL Image
        import io
        from PIL import Image
        img_data = pix.tobytes("png")
        img = Image.open(io.BytesIO(img_data)).convert("RGB")

        # Process with UniMERNet
        device = next(self.formula_recognizer.parameters()).device
        image_tensor = self.formula_processor(img).unsqueeze(0).to(device)

        with torch.no_grad():
            output = self.formula_recognizer.generate({"image": image_tensor})

        latex_code = output["pred_str"][0]
        return latex_code

    def _latex_to_mathml(self, latex_code):
        """
        Convert LaTeX to MathML

        Args:
            latex_code: LaTeX string

        Returns:
            str: MathML string
        """
        try:
            from latex2mathml.converter import convert
            mathml = convert(latex_code)
            return mathml
        except Exception as e:
            print(f"[Warning] LaTeX to MathML conversion failed: {e}")
            print(f"  LaTeX: {latex_code}")
            return None

    def convert(self, pdf_path, docx_path, **kwargs):
        """
        Convert PDF to DOCX with hybrid approach

        Args:
            pdf_path: Input PDF file
            docx_path: Output DOCX file
            **kwargs: Additional parameters for pdf2docx

        Supported kwargs:
            - use_yolo_images: bool (default True) - Use YOLO for image detection
            - yolo_conf: float (default 0.25) - YOLO confidence threshold
            - yolo_imgsz: int (default 1024) - YOLO image size
            - ... (other pdf2docx parameters)
        """
        use_yolo = kwargs.pop('use_yolo_images', True)

        # Update YOLO parameters if provided
        if 'yolo_conf' in kwargs:
            self.yolo_detector.conf = kwargs.pop('yolo_conf')
        if 'yolo_imgsz' in kwargs:
            self.yolo_detector.imgsz = kwargs.pop('yolo_imgsz')

        print(f"\n{'='*80}")
        print(f"Hybrid Conversion: {pdf_path}")
        print(f"  - Images & Formulas: {'DocLayout-YOLO + UniMERNet' if use_yolo else 'pdf2docx'}")
        print(f"  - Text/Tables: pdf2docx")
        print(f"{'='*80}\n")

        if use_yolo:
            # Pre-detect all figures and formulas with YOLO
            print("[YOLO] Pre-detecting figures and formulas in all pages...")
            pdf_doc = fitz.open(pdf_path)

            for page_num in range(len(pdf_doc)):
                page = pdf_doc[page_num]
                self.yolo_detector.detect_page_regions(page, page_num)

            pdf_doc.close()
            print()

            # Inject YOLO detector into pdf2docx
            self._inject_yolo_to_pdf2docx(pdf_path)

        # Convert with pdf2docx
        print("[pdf2docx] Converting with pdf2docx...")
        cv = Converter(pdf_path)

        # Set default parameters for better quality
        default_params = {
            'start': 0,
            'end': None,
        }
        default_params.update(kwargs)

        cv.convert(docx_path, **default_params)
        cv.close()

        # Insert formulas if any were detected
        if use_yolo:
            self._insert_formulas_to_docx(pdf_path, docx_path)

        print(f"\n{'='*80}")
        print(f"Conversion completed!")
        print(f"Output: {docx_path}")
        print(f"{'='*80}\n")

    def _insert_formulas_to_docx(self, pdf_path, docx_path):
        """
        Insert MathML formulas into the generated DOCX file

        Args:
            pdf_path: Original PDF file
            docx_path: Generated DOCX file
        """
        # Check if there are any formulas detected
        has_formulas = False
        for regions in self.yolo_detector.detected_regions.values():
            if regions.get('formulas'):
                has_formulas = True
                break

        if not has_formulas:
            return

        print("\n[Formula] Processing formulas...")

        # Open PDF
        pdf_doc = fitz.open(pdf_path)

        # Recognize all formulas
        formula_data = {}  # page_num -> [(bbox, latex, mathml), ...]

        for page_num, regions in self.yolo_detector.detected_regions.items():
            formulas = regions.get('formulas', [])
            if not formulas:
                continue

            page = pdf_doc[page_num]
            page_formulas = []

            for formula_idx, formula in enumerate(formulas):
                bbox = formula['bbox']

                # Recognize formula
                print(f"[Formula] Page {page_num + 1}: Recognizing formula at {bbox}...")
                latex = self._recognize_formula(page, bbox)
                print(f"  LaTeX: {latex}")

                # Convert to MathML
                mathml = self._latex_to_mathml(latex)

                if mathml:
                    formula_info = {
                        'bbox': bbox,
                        'latex': latex,
                        'mathml': mathml
                    }

                    # Preserve equation number if detected
                    if 'equation_number' in formula:
                        formula_info['equation_number'] = formula['equation_number']
                        print(f"  Equation number: {formula['equation_number']}")

                    page_formulas.append(formula_info)
                    print(f"  ✓ Converted to MathML")
                else:
                    print(f"  ✗ MathML conversion failed")

            if page_formulas:
                formula_data[page_num] = page_formulas

        pdf_doc.close()

        # Insert formulas into DOCX
        if formula_data:
            print(f"\n[Formula] Inserting {sum(len(f) for f in formula_data.values())} formulas into DOCX...")
            self._replace_formula_placeholders(docx_path, formula_data)
            print(f"[Formula] Formulas inserted successfully\n")

    def _replace_formula_placeholders(self, docx_path, formula_data):
        """
        Replace formula placeholder text with actual MathML formulas in DOCX file

        Args:
            docx_path: Generated DOCX file
            formula_data: Dict of {page_num: [{'bbox': [...], 'latex': str, 'mathml': str}]}
        """
        from docx import Document
        from docx.enum.text import WD_ALIGN_PARAGRAPH
        import re

        # Open document
        doc = Document(docx_path)

        # Create a mapping of placeholder IDs to formula data
        formula_map = {}
        for page_num, formulas in formula_data.items():
            for formula_idx, formula_info in enumerate(formulas):
                placeholder_id = f"FORMULA_PLACEHOLDER_{page_num}_{formula_idx}"
                formula_map[placeholder_id] = formula_info

        print(f"[Formula] Searching for {len(formula_map)} placeholder texts in document...")

        replaced_count = 0

        # Iterate through all paragraphs
        for para_idx, para in enumerate(doc.paragraphs):
            para_text = para.text

            # Look for formula placeholder pattern
            for placeholder_id, formula_info in formula_map.items():
                if placeholder_id in para_text:
                    mathml = formula_info['mathml']
                    latex = formula_info['latex']
                    equation_number = formula_info.get('equation_number', None)

                    print(f"  [Formula] Found placeholder '{placeholder_id}' at paragraph {para_idx}, replacing...")

                    try:
                        # Extract text before and after the placeholder
                        placeholder_pos = para_text.find(placeholder_id)
                        text_before = para_text[:placeholder_pos].strip()
                        text_after = para_text[placeholder_pos + len(placeholder_id):].strip()

                        # Remove equation number from text_after if present
                        # (equation number is already in the table, no need to duplicate)
                        if equation_number and text_after:
                            import re
                            # Remove all occurrences of equation number pattern
                            pattern = re.escape(equation_number)
                            # Remove pattern with surrounding whitespace/tabs (global replace)
                            text_after = re.sub(r'[\s\t]*' + pattern + r'[\s\t]*', '', text_after)
                            # Clean up any remaining tabs or excessive whitespace
                            text_after = re.sub(r'\t+', ' ', text_after).strip()
                            # If nothing left, set to empty
                            if not text_after or text_after.isspace():
                                text_after = ''

                        # Strip MathML declaration if present
                        if mathml.startswith('<?xml'):
                            mathml = mathml[mathml.index('?>') + 2:].strip()

                        # Create OMML element from MathML
                        math_run = self._create_omml_from_mathml(mathml)

                        if math_run:
                            # Use a single-row, 3-column invisible table for perfect alignment
                            # This is the standard professional method in Word for centered formula + right-aligned number
                            from docx.shared import Pt, Inches
                            from docx.oxml import OxmlElement
                            from docx.oxml.ns import qn
                            from docx.enum.text import WD_ALIGN_PARAGRAPH

                            # Create a table structure in XML (1 row, 3 columns)
                            # Column 1: Empty (balance)
                            # Column 2: Formula (centered)
                            # Column 3: Number (right-aligned)

                            tbl = OxmlElement('w:tbl')

                            # Table properties - no borders, full text width
                            tblPr = OxmlElement('w:tblPr')

                            # Table width = 100% of text width (between margins)
                            tblW = OxmlElement('w:tblW')
                            tblW.set(qn('w:type'), 'pct')
                            tblW.set(qn('w:w'), '5000')  # 100% = 5000 in pct
                            tblPr.append(tblW)

                            # Table indent = 0 (align with text left margin)
                            tblInd = OxmlElement('w:tblInd')
                            tblInd.set(qn('w:w'), '0')
                            tblInd.set(qn('w:type'), 'dxa')
                            tblPr.append(tblInd)

                            # No borders
                            tblBorders = OxmlElement('w:tblBorders')
                            for border_name in ['top', 'left', 'bottom', 'right', 'insideH', 'insideV']:
                                border = OxmlElement(f'w:{border_name}')
                                border.set(qn('w:val'), 'none')
                                border.set(qn('w:sz'), '0')
                                border.set(qn('w:space'), '0')
                                border.set(qn('w:color'), 'auto')
                                tblBorders.append(border)
                            tblPr.append(tblBorders)

                            # Table layout: fixed
                            tblLayout = OxmlElement('w:tblLayout')
                            tblLayout.set(qn('w:type'), 'fixed')
                            tblPr.append(tblLayout)

                            # No cell spacing
                            tblCellSpacing = OxmlElement('w:tblCellSpacing')
                            tblCellSpacing.set(qn('w:w'), '0')
                            tblCellSpacing.set(qn('w:type'), 'dxa')
                            tblPr.append(tblCellSpacing)

                            tbl.append(tblPr)

                            # Table grid - 3 columns with equal width
                            tblGrid = OxmlElement('w:tblGrid')
                            for _ in range(3):
                                gridCol = OxmlElement('w:gridCol')
                                tblGrid.append(gridCol)
                            tbl.append(tblGrid)

                            # Create the single row
                            tr = OxmlElement('w:tr')

                            # Row height auto
                            trPr = OxmlElement('w:trPr')
                            trHeight = OxmlElement('w:trHeight')
                            trHeight.set(qn('w:val'), '0')
                            trHeight.set(qn('w:hRule'), 'auto')
                            trPr.append(trHeight)
                            tr.append(trPr)

                            # Cell 1: Empty (for balance) - 5%
                            tc1 = OxmlElement('w:tc')
                            tcPr1 = OxmlElement('w:tcPr')
                            tcW1 = OxmlElement('w:tcW')
                            tcW1.set(qn('w:type'), 'pct')
                            tcW1.set(qn('w:w'), '250')  # 5% (5000/100*5 = 250)
                            tcPr1.append(tcW1)
                            tc1.append(tcPr1)
                            p1 = OxmlElement('w:p')
                            tc1.append(p1)
                            tr.append(tc1)

                            # Cell 2: Formula (centered) - 80%
                            tc2 = OxmlElement('w:tc')
                            tcPr2 = OxmlElement('w:tcPr')
                            tcW2 = OxmlElement('w:tcW')
                            tcW2.set(qn('w:type'), 'pct')
                            tcW2.set(qn('w:w'), '4000')  # 80% (5000/100*80 = 4000)
                            tcPr2.append(tcW2)
                            tc2.append(tcPr2)

                            p2 = OxmlElement('w:p')
                            # Center alignment for formula
                            pPr2 = OxmlElement('w:pPr')
                            jc2 = OxmlElement('w:jc')
                            jc2.set(qn('w:val'), 'center')
                            pPr2.append(jc2)
                            p2.append(pPr2)
                            # Add the math run to this paragraph
                            p2.append(math_run)
                            tc2.append(p2)
                            tr.append(tc2)

                            # Cell 3: Equation number (right-aligned) - 15%
                            tc3 = OxmlElement('w:tc')
                            tcPr3 = OxmlElement('w:tcPr')
                            tcW3 = OxmlElement('w:tcW')
                            tcW3.set(qn('w:type'), 'pct')
                            tcW3.set(qn('w:w'), '750')  # 15% (5000/100*15 = 750)
                            tcPr3.append(tcW3)
                            tc3.append(tcPr3)

                            p3 = OxmlElement('w:p')
                            if equation_number:
                                # Right alignment for equation number
                                pPr3 = OxmlElement('w:pPr')
                                jc3 = OxmlElement('w:jc')
                                jc3.set(qn('w:val'), 'right')
                                pPr3.append(jc3)
                                p3.append(pPr3)

                                # Add the equation number as text run
                                r3 = OxmlElement('w:r')
                                rPr3 = OxmlElement('w:rPr')
                                sz3 = OxmlElement('w:sz')
                                sz3.set(qn('w:val'), '22')  # 11pt
                                rPr3.append(sz3)
                                r3.append(rPr3)

                                t3 = OxmlElement('w:t')
                                t3.text = equation_number
                                r3.append(t3)
                                p3.append(r3)

                                print(f"    ✓ Added equation number: {equation_number}")
                            tc3.append(p3)
                            tr.append(tc3)

                            # Add row to table
                            tbl.append(tr)

                            # Now insert the table into the document
                            parent = para._element.getparent()
                            para_index = parent.index(para._element)

                            if text_before:
                                print(f"    [Info] Preserving text before formula: {text_before[:50]}...")
                                # Keep the paragraph with text_before
                                para.clear()
                                para.add_run(text_before)
                                # Insert table after this paragraph
                                parent.insert(para_index + 1, tbl)
                            else:
                                # Replace the paragraph with the table
                                para.clear()
                                parent.insert(para_index, tbl)
                                parent.remove(para._element)

                            # If there's text after the placeholder, add it as a new paragraph after the table
                            if text_after:
                                print(f"    [Info] Preserving text after formula: {text_after[:50]}...")
                                new_para = OxmlElement('w:p')
                                new_run = OxmlElement('w:r')
                                new_text = OxmlElement('w:t')
                                new_text.text = text_after
                                new_run.append(new_text)
                                new_para.append(new_run)

                                # Insert after table
                                if text_before:
                                    # Table was inserted at para_index + 1, so insert text_after at para_index + 2
                                    parent.insert(para_index + 2, new_para)
                                else:
                                    # Table was inserted at para_index, so insert text_after at para_index + 1
                                    parent.insert(para_index + 1, new_para)

                            replaced_count += 1
                            print(f"    ✓ Replaced with: {latex[:60]}...")
                        else:
                            print(f"    ✗ Failed to create OMML for: {latex[:60]}...")

                    except Exception as e:
                        print(f"    ✗ Error replacing formula: {e}")
                        import traceback
                        traceback.print_exc()

                    break  # Move to next paragraph

        print(f"[Formula] Replaced {replaced_count} out of {len(formula_map)} formulas")

        if replaced_count < len(formula_map):
            print(f"[Warning] {len(formula_map) - replaced_count} placeholders not found in document")

        # Save document
        doc.save(docx_path)

        # Post-process the DOCX file to remove whitespace from formulas
        self._cleanup_formula_whitespace(docx_path)

        # Process inline math symbols in text (always enabled when formulas are detected)
        self._process_inline_math(docx_path)

    def _process_inline_math(self, docx_path):
        """
        Process inline math symbols and expressions in paragraph text and convert to OMML

        Args:
            docx_path: Path to DOCX file
        """
        from docx import Document
        from docx.oxml import OxmlElement, parse_xml
        from docx.oxml.ns import qn
        import re

        print("\n[InlineMath] Processing inline math symbols and expressions...")

        doc = Document(docx_path)

        # Define math symbol patterns
        math_symbols = {
            '∈': ('element-of', 'in'),
            '∉': ('not-element-of', 'notin'),
            '⊂': ('subset', 'subset'),
            '⊃': ('superset', 'supset'),
            '∩': ('intersection', 'cap'),
            '∪': ('union', 'cup'),
            '∀': ('for-all', 'forall'),
            '∃': ('exists', 'exists'),
            '∇': ('nabla', 'nabla'),
            '∂': ('partial', 'partial'),
            '∫': ('integral', 'int'),
            '∑': ('sum', 'sum'),
            '∏': ('product', 'prod'),
            '√': ('square-root', 'sqrt'),
            '±': ('plus-minus', 'pm'),
            '×': ('times', 'times'),
            '÷': ('divide', 'div'),
            '≤': ('less-equal', 'leq'),
            '≥': ('greater-equal', 'geq'),
            '≠': ('not-equal', 'neq'),
            '≈': ('approximately', 'approx'),
            '∞': ('infinity', 'infty'),
            '∼': ('tilde', 'sim'),
            '→': ('arrow-right', 'rightarrow'),
            '←': ('arrow-left', 'leftarrow'),
            # Greek letters
            'α': ('alpha', 'alpha'),
            'β': ('beta', 'beta'),
            'γ': ('gamma', 'gamma'),
            'δ': ('delta', 'delta'),
            'ε': ('epsilon', 'epsilon'),
            'ζ': ('zeta', 'zeta'),
            'η': ('eta', 'eta'),
            'θ': ('theta', 'theta'),
            'ι': ('iota', 'iota'),
            'κ': ('kappa', 'kappa'),
            'λ': ('lambda', 'lambda'),
            'μ': ('mu', 'mu'),
            'ν': ('nu', 'nu'),
            'ξ': ('xi', 'xi'),
            'ο': ('omicron', 'omicron'),
            'π': ('pi', 'pi'),
            'ρ': ('rho', 'rho'),
            'σ': ('sigma', 'sigma'),
            'τ': ('tau', 'tau'),
            'υ': ('upsilon', 'upsilon'),
            'φ': ('phi', 'phi'),
            'χ': ('chi', 'chi'),
            'ψ': ('psi', 'psi'),
            'ω': ('omega', 'omega'),
        }

        # Count how many paragraphs we process
        processed_count = 0
        expression_count = 0
        subscript_count = 0
        symbol_count = 0

        for para in doc.paragraphs:
            text = para.text

            # Check if paragraph contains any math symbols or expressions
            has_math = any(symbol in text for symbol in math_symbols.keys())

            # Also check for math expression patterns
            has_expressions = bool(re.search(r'[a-zA-Z]+\([^)]*[θαβγδεζηικλμνξοπρστυφχψω∈×∼][^)]*\)', text))

            # Check for subscript patterns
            has_subscripts = bool(re.search(r'\b[a-zA-Z][0-9t]\b', text))

            if has_math or has_expressions or has_subscripts:
                # Process this paragraph
                expr_count, sym_count = self._convert_inline_math_in_paragraph(para, math_symbols)
                # expr_count now includes both expressions and subscripts
                # We'll count subscripts separately in the detailed output
                expression_count += expr_count
                symbol_count += sym_count
                processed_count += 1

        print(f"[InlineMath] Processed {processed_count} paragraphs")
        print(f"[InlineMath] - {expression_count} math expressions (including subscripts)")
        print(f"[InlineMath] - {symbol_count} individual symbols")

        # Save document
        doc.save(docx_path)

    def _convert_inline_math_in_paragraph(self, para, math_symbols):
        """
        Convert inline math symbols and expressions in a paragraph to OMML format

        Args:
            para: docx.paragraph.Paragraph object
            math_symbols: Dictionary of symbol -> (name, latex) mappings

        Returns:
            tuple: (expression_count, symbol_count)
        """
        from docx.oxml import OxmlElement, parse_xml
        from docx.oxml.ns import qn
        import re

        # Get paragraph text
        text = para.text

        math_items = []  # List of (start_pos, end_pos, text, type)

        # Step 1: Find all complete math expressions
        # Define multiple expression patterns

        # Pattern 1a: Distribution expressions like "x0 ∼N(0, I)"
        # Match: variable + ∼ + function with parentheses
        dist_func_pattern = r'\b[a-zA-Z][a-zA-Z0-9]*\s*∼\s*[A-Z][a-zA-Z]*\([^)]{0,50}\)'
        for match in re.finditer(dist_func_pattern, text):
            start, end = match.span()
            expr_text = match.group()
            math_items.append((start, end, expr_text, 'expression'))

        # Pattern 1b: Set membership with ranges like "t ∈[0, 1]"
        # Match: variable + ∈ + [range]
        set_range_pattern = r'\b[a-zA-Z][a-zA-Z0-9]*\s*∈\s*\[[^\]]{0,50}\]'
        for match in re.finditer(set_range_pattern, text):
            start, end = match.span()
            expr_text = match.group()
            # Check if not already covered
            if not any(s <= start < e for s, e, _, _ in math_items):
                math_items.append((start, end, expr_text, 'expression'))

        # Pattern 1c: Set membership with space like "x ∈R1+T/4×H/8×W/8"
        # Match: variable + ∈ + complex expression with math symbols
        set_expr_pattern = r'\b[a-zA-Z][a-zA-Z0-9]*\s*∈\s*[A-Z][A-Za-z0-9\+\-\*/×÷\{\}]{2,60}'
        for match in re.finditer(set_expr_pattern, text):
            start, end = match.span()
            expr_text = match.group()
            # Check if not already covered
            if not any(s <= start < e for s, e, _, _ in math_items):
                # Trim at common delimiters
                expr_text = re.split(r'[,;\.]\s', expr_text)[0]
                end = start + len(expr_text)
                math_items.append((start, end, expr_text, 'expression'))

        # Pattern 2: Function calls with math symbols like "u(xt, ctxt, t; θ)"
        func_pattern = r'\b[a-zA-Z_]+\([^)]{0,100}[θαβγδεζηικλμνξοπρστυφχψω∈×∼∀∃∇∂∫∑∏±÷≤≥≠≈∞→←][^)]{0,100}\)'
        for match in re.finditer(func_pattern, text):
            start, end = match.span()
            expr_text = match.group()
            # Check if not already covered
            if not any(s <= start < e for s, e, _, _ in math_items):
                math_items.append((start, end, expr_text, 'expression'))

        # Pattern 3: Equations with math symbols like "xt = tx1 + (1-t)x0"
        # Match: variable = expression with math symbols (but stop at sentence boundaries)
        eq_pattern = r'\b[a-zA-Z][a-zA-Z0-9]*\s*=\s*[^\.,;]{1,80}[θαβγδεζηικλμνξοπρστυφχψω∈×∼∀∃∇∂∫∑∏±÷≤≥≠≈∞→←−][^\.,;]{0,20}'
        for match in re.finditer(eq_pattern, text):
            start, end = match.span()
            expr_text = match.group()
            # Check if not already covered and not too long
            if not any(s <= start < e for s, e, _, _ in math_items) and len(expr_text) < 100:
                # Trim at sentence boundaries or common words
                expr_text = re.split(r'[,;\.]\s+(?=[A-Z])', expr_text)[0]  # Stop at ". Word"

                # Stop before common connecting words
                common_words = r'\b(where|and|or|is|are|was|were|represents|denotes|indicates|means|such|that|which|with|for|from|to|in|on|at|by|as)\b'
                match_word = re.search(common_words, expr_text)
                if match_word:
                    expr_text = expr_text[:match_word.start()].rstrip()

                expr_text = expr_text.rstrip('.,;: ')
                end = start + len(expr_text)
                if len(expr_text) > 0:
                    math_items.append((start, end, expr_text, 'expression'))

        # Step 2: Find subscript patterns like x0, x1, xt, ctxt
        # Pattern: single letter + digits/letters (subscript)
        # Be careful not to match technical terms like HTML5, T5-small, etc.

        # Pattern 2a: Single letter + numbers (x0, x1, x2)
        subscript_pattern = r'\b([a-zA-Zα-ωΑ-Ω])([0-9]+)\b'
        for match in re.finditer(subscript_pattern, text):
            start, end = match.span()
            # Check if not already covered
            if not any(s <= start < e for s, e, _, _ in math_items):
                base = match.group(1)
                subscript = match.group(2)

                # Skip common technical terms
                full_text = match.group(0)
                if full_text.lower() in ['t5', 'h264', 'h265', 'utf8', 'base64', 'md5', 'sha256']:
                    continue

                math_items.append((start, end, f"{base}_{subscript}", 'subscript'))

        # Pattern 2b: Single letter + letter (xt means x_t, vt means v_t)
        subscript_letter_pattern = r'\b([a-zA-Z])([a-z])\b'
        for match in re.finditer(subscript_letter_pattern, text):
            start, end = match.span()
            # Check if not already covered
            if not any(s <= start < e for s, e, _, _ in math_items):
                base = match.group(1)
                subscript = match.group(2)
                full_text = match.group(0)

                # Only if both are lowercase or first is uppercase
                # Skip common words like "is", "it", "at", "in", "on", "to", etc.
                if full_text in ['is', 'it', 'at', 'in', 'on', 'to', 'of', 'or', 'an', 'as', 'be', 'by', 'he', 'we', 'if', 'so', 'up', 'no']:
                    continue

                # Only if looks like a variable (short and mathematical context)
                if len(full_text) == 2:
                    math_items.append((start, end, f"{base}_{subscript}", 'subscript'))

        # Pattern 2c: Multi-letter variable + letter subscript (ctxt means c_{txt})
        multi_subscript_pattern = r'\b([a-z]{1,3})([a-z]{1,3})\b'
        for match in re.finditer(multi_subscript_pattern, text):
            start, end = match.span()
            full_text = match.group(0)

            # Only process specific known patterns from the context
            if full_text in ['ctxt']:  # Add more as needed
                # Check if not already covered
                if not any(s <= start < e for s, e, _, _ in math_items):
                    base = match.group(1)
                    subscript = match.group(2)
                    math_items.append((start, end, f"{base}_{subscript}", 'subscript'))

        # Step 3: Find individual math symbols (not already in expressions)
        for symbol in math_symbols.keys():
            pos = 0
            while True:
                pos = text.find(symbol, pos)
                if pos == -1:
                    break

                # Check if this symbol is already part of an expression
                in_expression = False
                for item_start, item_end, _, item_type in math_items:
                    if item_start <= pos < item_end:
                        in_expression = True
                        break

                if not in_expression:
                    math_items.append((pos, pos + len(symbol), symbol, 'symbol'))

                pos += 1

        # If no math items, return
        if not math_items:
            return 0, 0

        # Sort by position
        math_items.sort(key=lambda x: x[0])

        # Remove overlapping items (keep the first one)
        filtered_items = []
        last_end = -1
        for start, end, content, item_type in math_items:
            if start >= last_end:
                filtered_items.append((start, end, content, item_type))
                last_end = end

        # Count items
        expression_count = sum(1 for _, _, _, t in filtered_items if t == 'expression')
        subscript_count = sum(1 for _, _, _, t in filtered_items if t == 'subscript')
        symbol_count = sum(1 for _, _, _, t in filtered_items if t == 'symbol')

        # Build new paragraph content with math runs
        parent = para._element.getparent()
        para_index = parent.index(para._element)

        # Create new paragraph element
        new_para = OxmlElement('w:p')

        # Copy paragraph properties if any
        pPr = para._element.find(qn('w:pPr'))
        if pPr is not None:
            new_para.append(pPr)

        # Build runs: text segments + math runs
        last_pos = 0
        for start, end, content, item_type in filtered_items:
            # Add text before this math item
            if start > last_pos:
                text_before = text[last_pos:start]
                text_run = OxmlElement('w:r')
                text_elem = OxmlElement('w:t')
                text_elem.set(qn('xml:space'), 'preserve')
                text_elem.text = text_before
                text_run.append(text_elem)
                new_para.append(text_run)

            # Add math run
            if item_type == 'expression':
                math_run = self._create_inline_math_expression(content)
            elif item_type == 'subscript':
                math_run = self._create_subscript_math(content)
            else:  # symbol
                math_run = self._create_inline_math_run(content)

            if math_run is not None:
                new_para.append(math_run)

            last_pos = end

        # Add remaining text
        if last_pos < len(text):
            text_after = text[last_pos:]
            text_run = OxmlElement('w:r')
            text_elem = OxmlElement('w:t')
            text_elem.set(qn('xml:space'), 'preserve')
            text_elem.text = text_after
            text_run.append(text_elem)
            new_para.append(text_run)

        # Replace old paragraph with new one
        parent.insert(para_index, new_para)
        parent.remove(para._element)

        return expression_count + subscript_count, symbol_count

    def _create_inline_math_expression(self, expr_text):
        """
        Create an inline math run (OMML) for a complete expression like u(xt, ctxt, t; θ)

        Args:
            expr_text: Expression string like "u(xt, ctxt, t; θ)"

        Returns:
            OxmlElement (w:r with m:oMath) or None
        """
        from docx.oxml import OxmlElement, parse_xml
        from docx.oxml.ns import qn
        import re

        try:
            # Build OMML content by processing the expression text
            # We need to identify and convert subscripts within the expression

            omml_parts = []
            i = 0
            text_len = len(expr_text)

            while i < text_len:
                # Try to match subscript patterns at current position
                remaining = expr_text[i:]

                # Pattern 1: Multi-letter + multi-letter (ctxt -> c_{txt})
                # Must check this FIRST before single letter patterns
                match = re.match(r'^(c)(txt)(?![a-z])', remaining)
                if match:
                    base = match.group(1)
                    subscript = match.group(2)
                    # Create subscript OMML
                    omml_parts.append(
                        f'<m:sSub>'
                        f'<m:e><m:r><m:t>{base}</m:t></m:r></m:e>'
                        f'<m:sub><m:r><m:t>{subscript}</m:t></m:r></m:sub>'
                        f'</m:sSub>'
                    )
                    i += len(match.group(0))
                    continue

                # Pattern 2: Single letter + digits (x0, x1)
                match = re.match(r'^([a-zA-Zα-ωΑ-Ω])([0-9]+)', remaining)
                if match:
                    base = match.group(1)
                    subscript = match.group(2)
                    # Create subscript OMML
                    omml_parts.append(
                        f'<m:sSub>'
                        f'<m:e><m:r><m:t>{base}</m:t></m:r></m:e>'
                        f'<m:sub><m:r><m:t>{subscript}</m:t></m:r></m:sub>'
                        f'</m:sSub>'
                    )
                    i += len(match.group(0))
                    continue

                # Pattern 3: Single letter + single letter (xt, vt)
                # But be careful with common words
                match = re.match(r'^([a-zA-Z])([a-z])(?![a-z])', remaining)
                if match:
                    base = match.group(1)
                    subscript = match.group(2)
                    full = match.group(0)

                    # Skip common words
                    if full not in ['is', 'it', 'at', 'in', 'on', 'to', 'of', 'or', 'an', 'as', 'be', 'by']:
                        # Create subscript OMML
                        omml_parts.append(
                            f'<m:sSub>'
                            f'<m:e><m:r><m:t>{base}</m:t></m:r></m:e>'
                            f'<m:sub><m:r><m:t>{subscript}</m:t></m:r></m:sub>'
                            f'</m:sSub>'
                        )
                        i += len(match.group(0))
                        continue

                # No subscript pattern matched, add single character as text
                char = expr_text[i]
                omml_parts.append(f'<m:r><m:t>{char}</m:t></m:r>')
                i += 1

            # Combine all parts
            omml_content = ''.join(omml_parts)

            # Remove whitespace between tags
            omml_content = re.sub(r'>\s+<', '><', omml_content)

            # Create OMML run
            omml_xml = (
                '<w:r xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" '
                'xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math">'
                f'<m:oMath>{omml_content}</m:oMath>'
                '</w:r>'
            )

            omml_elem = parse_xml(omml_xml)
            return omml_elem

        except Exception as e:
            print(f"[Warning] Failed to create inline math expression for '{expr_text}': {e}")
            import traceback
            traceback.print_exc()
            return None

    def _create_subscript_math(self, subscript_text):
        """
        Create an inline math run (OMML) for subscript like x_0, v_t, c_{txt}

        Args:
            subscript_text: Subscript string in format "base_subscript" (e.g., "x_0", "v_t")

        Returns:
            OxmlElement (w:r with m:oMath containing m:sSub) or None
        """
        from docx.oxml import parse_xml
        import re

        try:
            # Parse the subscript notation
            parts = subscript_text.split('_')
            if len(parts) != 2:
                return self._create_inline_math_run(subscript_text)

            base, subscript = parts

            # Build OMML subscript structure
            # <m:sSub> = subscript element
            #   <m:e> = base element
            #   <m:sub> = subscript element

            omml_content = (
                f'<m:sSub>'
                f'<m:e><m:r><m:t>{base}</m:t></m:r></m:e>'
                f'<m:sub><m:r><m:t>{subscript}</m:t></m:r></m:sub>'
                f'</m:sSub>'
            )

            # Remove whitespace between tags
            omml_content = re.sub(r'>\s+<', '><', omml_content)

            # Create OMML run
            omml_xml = (
                '<w:r xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" '
                'xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math">'
                f'<m:oMath>{omml_content}</m:oMath>'
                '</w:r>'
            )

            omml_elem = parse_xml(omml_xml)
            return omml_elem

        except Exception as e:
            print(f"[Warning] Failed to create subscript math for '{subscript_text}': {e}")
            return None

    def _create_inline_math_run(self, symbol):
        """
        Create an inline math run (OMML) for a single symbol

        Args:
            symbol: Unicode math symbol string

        Returns:
            OxmlElement (w:r with m:oMath) or None
        """
        from docx.oxml import OxmlElement, parse_xml
        from docx.oxml.ns import qn

        try:
            # Create a simple OMML structure for the symbol
            # We'll use m:r (math run) with m:t (math text)
            omml_xml = (
                '<w:r xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" '
                'xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math">'
                '<m:oMath><m:r><m:t>' + symbol + '</m:t></m:r></m:oMath>'
                '</w:r>'
            )

            omml_elem = parse_xml(omml_xml)
            return omml_elem

        except Exception as e:
            print(f"[Warning] Failed to create inline math for symbol '{symbol}': {e}")
            return None

    def _cleanup_formula_whitespace(self, docx_path):
        """
        Post-process DOCX file to remove whitespace from math formulas

        Args:
            docx_path: Path to DOCX file
        """
        import zipfile
        import tempfile
        import os
        import re
        from lxml import etree

        try:
            # Create a temporary file
            temp_fd, temp_path = tempfile.mkstemp(suffix='.docx')
            os.close(temp_fd)

            # Open the DOCX as a ZIP file
            with zipfile.ZipFile(docx_path, 'r') as zip_in:
                with zipfile.ZipFile(temp_path, 'w', zipfile.ZIP_DEFLATED) as zip_out:
                    for item in zip_in.namelist():
                        data = zip_in.read(item)

                        # Process word/document.xml to clean formula whitespace
                        if item == 'word/document.xml':
                            # Parse XML
                            root = etree.fromstring(data)

                            # Find all m:oMath elements and remove whitespace
                            nsmap = {
                                'w': 'http://schemas.openxmlformats.org/wordprocessingml/2006/main',
                                'm': 'http://schemas.openxmlformats.org/officeDocument/2006/math'
                            }

                            for math_elem in root.xpath('.//m:oMath', namespaces=nsmap):
                                # Remove text content (whitespace) directly under m:oMath
                                if math_elem.text and not math_elem.text.strip():
                                    math_elem.text = None

                                # Remove whitespace tails from all descendants
                                for elem in math_elem.iter():
                                    if elem.tail and not elem.tail.strip():
                                        elem.tail = None

                            # Serialize back to bytes without pretty printing
                            data = etree.tostring(root, encoding='utf-8', xml_declaration=True)

                        # Write to new ZIP
                        zip_out.writestr(item, data)

            # Replace original file with cleaned version
            import shutil
            shutil.move(temp_path, docx_path)

            print(f"[Formula] Cleaned whitespace from formulas")

        except Exception as e:
            print(f"[Warning] Failed to cleanup formula whitespace: {e}")
            import traceback
            traceback.print_exc()
            # Clean up temp file if it exists
            if os.path.exists(temp_path):
                try:
                    os.remove(temp_path)
                except:
                    pass

    def _remove_blank_text_nodes(self, element):
        """
        Recursively remove blank text nodes from an XML element tree

        Args:
            element: lxml Element
        """
        # Remove blank text at element start
        if element.text and not element.text.strip():
            element.text = None

        # Process all children
        for child in element:
            # Remove blank tail of previous sibling
            if child.tail and not child.tail.strip():
                child.tail = None
            # Recursively process child
            self._remove_blank_text_nodes(child)

    def _mathml_to_omml(self, mathml_elem):
        """
        Convert MathML element to OMML (Office Math ML) element recursively

        Args:
            mathml_elem: lxml Element with MathML

        Returns:
            OMML element as XML string
        """
        from lxml import etree

        # Remove namespace for easier handling
        tag = mathml_elem.tag.split('}')[-1] if '}' in mathml_elem.tag else mathml_elem.tag

        # Handle different MathML elements
        if tag == 'math':
            # Root element - process children
            omml_parts = []
            for child in mathml_elem:
                omml_parts.append(self._mathml_to_omml(child))
            return ''.join(omml_parts)

        elif tag == 'mrow':
            # Group - just process children
            omml_parts = []
            for child in mathml_elem:
                omml_parts.append(self._mathml_to_omml(child))
            return ''.join(omml_parts)

        elif tag == 'msup':
            # Superscript: base^superscript
            children = list(mathml_elem)
            if len(children) >= 2:
                base = self._mathml_to_omml(children[0])
                sup = self._mathml_to_omml(children[1])
                return f'<m:sSup><m:e>{base}</m:e><m:sup>{sup}</m:sup></m:sSup>'
            return ''

        elif tag == 'msub':
            # Subscript: base_subscript
            children = list(mathml_elem)
            if len(children) >= 2:
                base = self._mathml_to_omml(children[0])
                sub = self._mathml_to_omml(children[1])
                return f'<m:sSub><m:e>{base}</m:e><m:sub>{sub}</m:sub></m:sSub>'
            return ''

        elif tag == 'msubsup':
            # Subscript and superscript: base_sub^sup
            children = list(mathml_elem)
            if len(children) >= 3:
                base = self._mathml_to_omml(children[0])
                sub = self._mathml_to_omml(children[1])
                sup = self._mathml_to_omml(children[2])
                return f'<m:sSubSup><m:e>{base}</m:e><m:sub>{sub}</m:sub><m:sup>{sup}</m:sup></m:sSubSup>'
            return ''

        elif tag == 'mfrac':
            # Fraction: numerator/denominator
            children = list(mathml_elem)
            if len(children) >= 2:
                num = self._mathml_to_omml(children[0])
                den = self._mathml_to_omml(children[1])
                return f'<m:f><m:num>{num}</m:num><m:den>{den}</m:den></m:f>'
            return ''

        elif tag in ['mi', 'mn', 'mo', 'mtext']:
            # Text elements: identifier, number, operator, text
            text = mathml_elem.text or ''
            # Escape XML special characters
            text = text.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')
            return f'<m:r><m:t>{text}</m:t></m:r>'

        else:
            # Unknown element - just process children and extract text
            text = ''.join(mathml_elem.itertext())
            text = text.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')
            return f'<m:r><m:t>{text}</m:t></m:r>'

    def _create_omml_from_mathml(self, mathml):
        """
        Create OMML (Office Math ML) run element from MathML

        Args:
            mathml: MathML string

        Returns:
            OxmlElement (w:r with m:oMath) or None
        """
        from docx.oxml import parse_xml
        from lxml import etree
        import re

        try:
            # Parse the MathML
            mathml_tree = etree.fromstring(mathml.encode('utf-8'))

            # Convert MathML to OMML
            omml_content = self._mathml_to_omml(mathml_tree).strip()

            # Remove all inter-tag whitespace (between > and <)
            omml_content = re.sub(r'>\s+<', '><', omml_content)

            # Create OMML run with math zone (no whitespace!)
            omml_xml = (
                '<w:r xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main" '
                'xmlns:m="http://schemas.openxmlformats.org/officeDocument/2006/math">'
                f'<m:oMath>{omml_content}</m:oMath>'
                '</w:r>'
            )

            # Parse with lxml first to clean whitespace
            parser = etree.XMLParser(remove_blank_text=True)
            lxml_elem = etree.fromstring(omml_xml.encode('utf-8'), parser=parser)

            # Convert back to string without pretty printing
            clean_xml = etree.tostring(lxml_elem, encoding='unicode')

            # Now parse with docx parser
            omml_elem = parse_xml(clean_xml)
            return omml_elem

        except Exception as e:
            print(f"[Warning] Failed to create OMML: {e}")
            import traceback
            traceback.print_exc()
            return None

    def _inject_yolo_to_pdf2docx(self, pdf_path):
        """
        Inject YOLO detected regions into pdf2docx's image extraction
        Also filter out text/shapes that overlap with YOLO detected regions

        This modifies pdf2docx's internal image extraction to use YOLO results
        """
        # Monkey-patch pdf2docx's ImagesExtractor
        from pdf2docx.image.ImagesExtractor import ImagesExtractor
        from pdf2docx.shape.Paths import Paths
        from pdf2docx.shape.Shapes import Shapes
        from pdf2docx.page.RawPageFitz import RawPageFitz

        # Save original methods
        original_extract_images = ImagesExtractor.extract_images
        original_paths_restore = Paths.restore
        original_shapes_restore = Shapes.restore
        original_to_shapes_and_images = Paths.to_shapes_and_images
        original_preprocess_text = RawPageFitz._preprocess_text

        yolo_detector = self.yolo_detector

        def _rect_overlap_ratio(rect1, rect2):
            """Calculate overlap ratio between two rectangles"""
            # Convert to fitz.Rect if needed
            r1 = fitz.Rect(rect1) if not isinstance(rect1, fitz.Rect) else rect1
            r2 = fitz.Rect(rect2) if not isinstance(rect2, fitz.Rect) else rect2

            # Get intersection
            intersection = r1 & r2
            if intersection.is_empty:
                return 0.0

            # Calculate areas
            intersection_area = intersection.width * intersection.height
            r1_area = r1.width * r1.height

            if r1_area == 0:
                return 0.0

            return intersection_area / r1_area

        def yolo_enhanced_extract_images(self, clip_image_res_ratio=3.0, **kwargs):
            """Enhanced image extraction using YOLO detections (figures only)"""

            page = self._page
            page_num = page.number

            # Get YOLO detected regions for this page
            regions = yolo_detector.detected_regions.get(page_num, {})
            yolo_figures = regions.get('figures', [])

            if yolo_figures:
                # Use YOLO detections for figures only
                images = []

                # Add figures
                for fig in yolo_figures:
                    bbox = fig['bbox']
                    rect = fitz.Rect(bbox)

                    # Extract image from PDF
                    mat = fitz.Matrix(clip_image_res_ratio, clip_image_res_ratio)
                    pix = page.get_pixmap(matrix=mat, clip=rect)

                    # Convert to pdf2docx format
                    if pix.colorspace.n > 3:
                        pix = fitz.Pixmap(fitz.csRGB, pix)

                    raw_dict = {
                        "type": 1,  # BlockType.IMAGE
                        "bbox": tuple(bbox),
                        "width": pix.width,
                        "height": pix.height,
                        "image": pix.tobytes(),
                    }

                    images.append(raw_dict)

                print(f"[YOLO] Page {page_num + 1}: Using {len(yolo_figures)} YOLO-detected figures")
                return images
            else:
                # Fall back to original pdf2docx image extraction
                result = original_extract_images(self, clip_image_res_ratio, **kwargs)
                if result:
                    print(f"[Debug] Page {page_num + 1}: extract_images() returned {len(result)} images (no YOLO)")
                return result

        def yolo_filtered_preprocess_text(self, **settings):
            """Filter out text blocks that overlap with YOLO regions, but insert formula placeholders"""
            # Get page number
            page_num = self.page_engine.number
            regions = yolo_detector.detected_regions.get(page_num, {})
            yolo_figures = regions.get('figures', [])

            # Call original method to get all text blocks
            text_blocks = original_preprocess_text(self, **settings)

            # If no YOLO figures or formulas, return original result
            yolo_formulas = regions.get('formulas', [])
            if not yolo_figures and not yolo_formulas:
                return text_blocks

            # First, detect equation numbers for formulas BEFORE filtering
            # Extract directly from PDF page instead of relying on text_blocks
            page = self.page_engine
            for formula_idx, formula in enumerate(yolo_formulas):
                bbox = formula['bbox']
                formula_rect = fitz.Rect(bbox)

                # Debug: print formula bbox
                print(f"[Debug] Page {page_num + 1}, Formula {formula_idx} bbox: {bbox}")

                # Try to find equation number by extracting text from area to the right of formula
                # Create a search rectangle to the right of the formula
                search_rect = fitz.Rect(
                    formula_rect.x1 - 5,  # Start slightly before formula end
                    formula_rect.y0 - 5,  # Expand vertically a bit
                    formula_rect.x1 + 100,  # Search 100 points to the right
                    formula_rect.y1 + 5
                )

                # Extract text from this region
                text_dict = page.get_textpage().extractDICT()
                equation_number = None

                for block in text_dict.get('blocks', []):
                    if block.get('type') != 0:  # Only text blocks
                        continue

                    block_bbox = fitz.Rect(block['bbox'])

                    # Check if this block intersects with our search rectangle
                    if not search_rect.intersects(block_bbox):
                        continue

                    # Extract text from this block
                    block_text = ''
                    for line in block.get('lines', []):
                        for span in line.get('spans', []):
                            block_text += span.get('text', '')

                    if not block_text.strip():
                        continue

                    print(f"[Debug]   Found text in search area: '{block_text}'")

                    # Check if it matches equation number pattern
                    import re
                    match = re.search(r'\(\s*\d+(\.\d+)?\s*\)', block_text)
                    if match:
                        equation_number = match.group(0)
                        print(f"[Formula] Detected equation number: '{equation_number}' for formula {formula_idx}")
                        break

                # Store in formula metadata
                if equation_number:
                    formula['equation_number'] = equation_number

            # Convert YOLO figure regions to Rect list (NOT formulas - we want to keep formula text areas)
            yolo_rects = [fitz.Rect(fig['bbox']) for fig in yolo_figures]

            # Filter out text blocks that overlap with YOLO figure regions (but not formula regions)
            # IMPROVED: Use higher threshold to preserve body text near figures
            filtered_blocks = []
            filtered_count = 0
            for block in text_blocks:
                # Check if this is a text block (type 0) vs image block (type 1)
                if block.get('type') != 0:
                    filtered_blocks.append(block)
                    continue

                block_bbox = fitz.Rect(block['bbox'])
                overlaps_with_figure = False
                for yolo_rect in yolo_rects:
                    overlap_ratio = _rect_overlap_ratio(block_bbox, yolo_rect)
                    # IMPROVED: Raised threshold from 0.5 to 0.85
                    # Only filter text that's almost completely inside a figure
                    # This preserves body text that's merely adjacent to figures
                    if overlap_ratio > 0.85:  # More than 85% overlap with figure
                        overlaps_with_figure = True
                        filtered_count += 1
                        break

                if not overlaps_with_figure:
                    filtered_blocks.append(block)

            # Insert formula placeholder text blocks
            for formula_idx, formula in enumerate(yolo_formulas):
                bbox = formula['bbox']
                formula_rect = fitz.Rect(bbox)
                placeholder_text = f"FORMULA_PLACEHOLDER_{page_num}_{formula_idx}"

                # Get equation number if detected earlier
                equation_number = formula.get('equation_number', None)

                # Create a text block for the placeholder
                placeholder_block = {
                    'type': 0,  # Text block
                    'bbox': tuple(bbox),
                    'lines': [{
                        'bbox': tuple(bbox),
                        'wmode': 0,
                        'dir': [1, 0],
                        'spans': [{
                            'bbox': tuple(bbox),
                            'text': placeholder_text,
                            'size': 12,
                            'flags': 0,
                            'font': 'Arial',
                            'color': 0
                        }]
                    }]
                }

                # Add equation number as a separate span if found
                if equation_number:
                    # Extend bbox to include number position (to the right)
                    extended_bbox = list(bbox)
                    extended_bbox[2] += 50  # Add some width for the number

                    placeholder_block['lines'][0]['spans'].append({
                        'bbox': tuple(extended_bbox),
                        'text': f" {equation_number}",  # Add space before number
                        'size': 12,
                        'flags': 0,
                        'font': 'Arial',
                        'color': 0
                    })

                filtered_blocks.append(placeholder_block)

            if filtered_count > 0:
                print(f"[Filter] Page {page_num + 1}: Filtered {filtered_count} text blocks overlapping with figures")
            if yolo_formulas:
                print(f"[Filter] Page {page_num + 1}: Inserted {len(yolo_formulas)} formula placeholders")

            return filtered_blocks

        def yolo_filtered_paths_restore(self, raws:list):
            """Completely skip paths for pages with YOLO detections"""
            # Get page number
            page_num = getattr(self.parent, 'number', 0)
            regions = yolo_detector.detected_regions.get(page_num, {})
            yolo_figures = regions.get('figures', [])
            yolo_formulas = regions.get('formulas', [])

            # If page has YOLO detections (figures or formulas), skip all path processing
            if yolo_figures or yolo_formulas:
                print(f"[Filter] Page {page_num + 1}: Disabled all path processing (using YOLO only)")
                return self  # Return empty paths
            else:
                # No YOLO detections, use original method
                return original_paths_restore(self, raws)

        def yolo_filtered_shapes_restore(self, raws:list):
            """Filter out shapes that overlap with YOLO detected regions"""
            # Get page number
            page_num = getattr(self.parent, 'number', 0)
            regions = yolo_detector.detected_regions.get(page_num, {})
            yolo_figures = regions.get('figures', [])
            yolo_formulas = regions.get('formulas', [])

            # If no YOLO figures or formulas, use original method
            if not yolo_figures and not yolo_formulas:
                return original_shapes_restore(self, raws)

            # Convert YOLO regions to Rect list (both figures and formulas)
            yolo_rects = [fitz.Rect(fig['bbox']) for fig in yolo_figures]
            yolo_rects.extend([fitz.Rect(formula['bbox']) for formula in yolo_formulas])

            # Filter shapes
            from pdf2docx.shape.Shape import Stroke, Fill, Hyperlink
            self.reset()
            rect = (0, 0, self.parent.width, self.parent.height)
            filtered_count = 0

            for raw in raws:
                # Distinguish specified type by key like `start`, `end` and `uri` (same as original)
                if 'start' in raw:
                    shape = Stroke(raw)
                elif 'uri' in raw:
                    shape = Hyperlink(raw)
                else:
                    shape = Fill(raw)

                # Ignore shape out of page
                if not shape.bbox.intersects(rect):
                    continue

                # Check if shape overlaps with any YOLO region
                overlaps = False
                for yolo_rect in yolo_rects:
                    overlap_ratio = _rect_overlap_ratio(shape.bbox, yolo_rect)
                    # IMPROVED: Keep at 90% for shapes (more conservative than text)
                    if overlap_ratio > 0.9:  # More than 90% overlap
                        overlaps = True
                        filtered_count += 1
                        break

                if not overlaps:
                    self.append(shape)

            if filtered_count > 0:
                print(f"[Filter] Page {page_num + 1}: Filtered {filtered_count} overlapping shapes")

            return self

        def yolo_filtered_to_shapes_and_images(self,
                                               min_svg_gap_dx:float=15,
                                               min_svg_gap_dy:float=15,
                                               min_w:float=2,
                                               min_h:float=2,
                                               clip_image_res_ratio:float=3.0):
            """Completely disable SVG image generation for pages with YOLO detections"""
            # Get page number
            page_num = getattr(self.parent, 'number', 0)
            regions = yolo_detector.detected_regions.get(page_num, {})
            yolo_figures = regions.get('figures', [])
            yolo_formulas = regions.get('formulas', [])

            # If page has YOLO detections (figures or formulas), disable all SVG/Path image generation
            if yolo_figures or yolo_formulas:
                # Only generate shapes (for tables/text styles), NO images
                iso_shapes = []
                if self.is_iso_oriented:
                    iso_shapes.extend(self.to_shapes())

                print(f"[Filter] Page {page_num + 1}: Disabled SVG/Path image generation (using YOLO only)")
                return iso_shapes, []  # Return empty images list
            else:
                # No YOLO detections, use original method
                return original_to_shapes_and_images(
                    self, min_svg_gap_dx, min_svg_gap_dy, min_w, min_h, clip_image_res_ratio)

        # Apply monkey-patches
        ImagesExtractor.extract_images = yolo_enhanced_extract_images
        RawPageFitz._preprocess_text = yolo_filtered_preprocess_text
        Paths.restore = yolo_filtered_paths_restore
        Paths.to_shapes_and_images = yolo_filtered_to_shapes_and_images
        # Shapes.restore = yolo_filtered_shapes_restore  # Disabled: may interfere with table parsing


def main():
    """Test the hybrid converter"""
    import argparse

    parser = argparse.ArgumentParser(description='Hybrid PDF to DOCX Converter')
    parser.add_argument('--pdf', required=True, help='Input PDF file')
    parser.add_argument('--output', help='Output DOCX file (default: input_hybrid.docx)')
    parser.add_argument('--yolo-model', default='weights/doclayout_yolo_docstructbench_imgsz1024.pt')
    parser.add_argument('--yolo-conf', type=float, default=0.25, help='YOLO confidence threshold')
    parser.add_argument('--yolo-imgsz', type=int, default=1024, help='YOLO image size')
    parser.add_argument('--no-yolo', action='store_true', help='Disable YOLO, use pdf2docx only')

    args = parser.parse_args()

    # Set default output path
    if args.output is None:
        args.output = args.pdf.replace('.pdf', '_hybrid.docx')

    # Create converter
    converter = HybridConverter(yolo_model_path=args.yolo_model)

    # Convert
    converter.convert(
        pdf_path=args.pdf,
        docx_path=args.output,
        use_yolo_images=not args.no_yolo,
        yolo_conf=args.yolo_conf,
        yolo_imgsz=args.yolo_imgsz
    )


if __name__ == "__main__":
    main()
